import corgidrp
import corgidrp.caldb as caldb
import corgidrp.walker as walker
import argparse


def step_1_initialize():
    """
    Initialize corgidrp and it's caldb again

    Returns: 
        caldb.CalDB: an instance of an initialized caldb object
    """
    corgidrp.create_config_dir()
    corgidrp.update_pipeline_settings()
    caldb.initialize()
    this_caldb = caldb.CalDB()
    return this_caldb

def step_2_load_cal(this_caldb, main_cal_dir):
    """

    Takes the initialized caldb and loads the calibration files from the main_cal_dir

    Args:
        this_caldb (caldb.CalDB): an instance of an initialized caldb object
        main_cal_dir (str): the path to the main calibration directory

    Returns:
        caldb.CalDB: an instance of a caldb object with the calibration files loaded
    """
    this_caldb.scan_dir_for_new_entries(main_cal_dir)
    return this_caldb

def step_3_process_data(input_filelist, cpgs_xml_filepath, outputdir, template=None):
     """
     
        Process the input file list by autodetecting a template, or accepting an optional template argument. 
        Data will be written out to the output directory.

        Args:
            input_filelist (list): a list of filepaths to the input files
            cpgs_xml_filepath (str): the path to the cpgs xml file
            outputdir (str): the path to the output directory
            template (str): the path to the template file (optional)

     
     """
     walker.walk_corgidrp(input_filelist, cpgs_xml_filepath, outputdir, template=template)


import os
import warnings
from astropy.io import fits
import numpy as np
from corgidrp.data import PyKLIPDataset, Image 
from pyklip.parallelized import klip_dataset
from pyklip.fakes import gaussfit2d, inject_planet
from scipy.ndimage import shift, rotate
from corgidrp.astrom import get_polar_dist, seppa2dxdy, seppa2xy
from corgidrp.fluxcal import phot_by_gauss2d_fit

def get_closest_psf(ct_calibration,cenx,ceny,dx,dy):
    """_summary_
    NOTE: CT excam locations have (0,0) as the bottom left corner 
      of the bottom left pixel

    TODO: Calculate subpixel shifts if star or PSF model aren't
        perfectly in the center of a pixel

    Args:
        ct_calibration (corgidrp.data.CoreThroughputCalibration): CT calibration object.
        cenx (float): x location of mask center, measured from the center of the bottom 
            left pixel of the 1024x1024 pixel science array.
        ceny (float): y location of mask center, measured from the center of the bottom 
            left pixel of the 1024x1024 pixel science array.
        dx (float): x separation from the mask center in pixels
        dy (float): y separation from the mask center in pixels

    Returns:
        np.array: 2D PSF model closest to the desired location.
    """

    # Shift so (0,0) is the center of the bottom left pixel
    x_arr = ct_calibration.ct_excam[0] - 0.5
    y_arr = ct_calibration.ct_excam[1] - 0.5

    rel_x_arr = x_arr - cenx
    rel_y_arr = y_arr - ceny

    dists = np.sqrt((rel_x_arr-dx)**2 + (rel_y_arr-dy)**2)

    arg_closest = np.argmin(dists)

    return ct_calibration.data[arg_closest] 


def inject_psf(frame_in, ct_calibration, amp, 
               sep_pix,pa_deg):
    """Injects a fake psf from the CT calibration object into a corgidrp Image with 
    the desired position and amplitude. 

    Args:
        frame_in (corgidrp.data.Image): 2D image to inject a fake signal into.
        ct_calibration (corgidrp.data.CoreThroughputCalibration): CT calibration object containing PSF samples.
        amp (float): peak pixel amplitude of psf to inject.
        sep_pix (float): separation from star in pixels to inject 
        pa_deg (float): position angle to inject (counterclockise from north/up)

    Returns: 
        corgidrp.data.Image: a copy of the input Image but with a fake PSF injected.
    """

    frame = frame_in.copy()

    # Get closest psf model
    frame_roll = frame.pri_hdr['ROLL']
    rel_pa = pa_deg - frame_roll
    dx,dy = seppa2dxdy(sep_pix,rel_pa)

    psf_model = get_closest_psf(ct_calibration,
                                frame.ext_hdr['STARLOCX'],
                                frame.ext_hdr['STARLOCY'],
                                dx,dy).copy() 

    # Scale counts
    peak_count = np.nanmax(psf_model)
    psf_model *= amp / peak_count

    # Assume PSF is centered in the data cutout for now
    model_shape = np.array(psf_model.shape)
    psf_cenyx_ind = (np.array(model_shape)/2 - 0.5).astype(int) 
    psf_cenyx_inframe = np.array([frame.ext_hdr['STARLOCY'],frame.ext_hdr['STARLOCX']]) + np.array([dy,dx])
    injected_psf_cenyx_ind = np.round(psf_cenyx_inframe).astype(int)
    startyx = injected_psf_cenyx_ind - psf_cenyx_ind
    endyx = startyx + model_shape

    inject_planet([frame.data], [[frame.ext_hdr['STARLOCX'], frame.ext_hdr['STARLOCY']]], [psf_model], [None], sep_pix, 0, thetas=[rel_pa + 90])

    psf_cenxy = [psf_cenyx_inframe[1],psf_cenyx_inframe[0]]
    return frame, psf_model, psf_cenxy


def measure_noise(frame, seps_pix, hw, klmode_index=None, cand_locs = []):
    """Calculates the noise (standard deviation of counts) of an 
        annulus at a given separation from the mask or star center.
        TODO: Correct for small sample statistics?
        TODO: Mask known off-axis sources.
    
    Args:
        frame (corgidrp.Image): Image containing data as well as "STARLOCX/Y" in header
        seps_pix (np.array of float): Separations (in pixels from specified center) at which to calculate 
            the noise level.
        hw (float): halfwidth of the annulus to use for noise calculation.
        klmode_index (int, optional): If provided, returns only the noise values for the KL mode with 
            the given index. I.e. klmode_index=0 would return only the values for the first KL mode 
            truncation choice.  If None (by default), all indices are returned.
        cand_locs (list of tuples, optional): Locations of known off-axis sources, so we can mask them. 
            This is a list of tuples (sep_pix,pa_degrees) for each source. Defaults to [].
        
    Returns: np.array: array of shape (number of separtions, number of KL modes) containing the annular noise.  If klmode_index 
        specified, the number of KL modes in the output array is 1.
    """
    cenx, ceny = (frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'])
    
    # Mask data outside the specified annulus
    y, x = np.indices(frame.data.shape[1:])
    sep_map = np.sqrt((y-ceny)**2 + (x-cenx)**2)
    sep_map3d = np.ones_like(frame.data) * sep_map

    stds = []
    for sep_pix in seps_pix:
        r_inner = sep_pix - hw
        r_outer = sep_pix + hw
        masked_data = np.where(np.logical_and(sep_map3d<r_outer,sep_map3d>r_inner), frame.data,np.nan)
        
        # import matplotlib.pyplot as plt
        # plt.imshow(masked_data[0],origin='lower')
        # plt.colorbar()
        # plt.title('Candlocs not masked')
        # plt.show()

        if len(cand_locs) > 0:
            for cand_loc in cand_locs:
                cand_x, cand_y = seppa2xy(*cand_loc,cenx,ceny)
                dists = np.sqrt((y-cand_y)**2 + (x-cand_x)**2)
    
                masked_data = np.where(dists > 5 * hw, masked_data.copy(),np.nan)
        
        # import matplotlib.pyplot as plt
        # plt.imshow(masked_data[0],origin='lower')
        # plt.colorbar()
        # plt.title('Candlocs masked')
        # plt.show()

        # Calculate standard deviation
        std = np.nanstd(masked_data,axis=(1,2))
        stds.append(std)

    stds_arr = np.array(stds)

    if klmode_index != None:
        return stds_arr[:,klmode_index]
    
    return stds_arr


def meas_klip_thrupt(sci_dataset_in,ref_dataset_in, # pre-psf-subtracted dataset
                     psfsub_dataset,
                     ct_calibration,
                     klip_params,
                     inject_snr = 5,
                     sep_spacing = 3.,
                     n_pas = 5,
                     seps = None, # in pixels from mask center
                     pas = None,
                     cand_locs = [], # list of tuples (sep_pix,pa_deg) of known off-axis source locations,
                     num_processes = None
                    ):
    """Measures the throughput of the KLIP algorithm via injection-recovery of fake off-axis sources. 

    Args:
        sci_dataset_in (corgidrp.data.Dataset): input dataset containing science observations
        ref_dataset_in (corgidrp.data.Dataset): input dataset containing reference observations 
            (set to None for ADI-only reductions)
        psfsub_dataset (corgidrp.data.Dataset): dataset containing PSF subtraction result
        ct_calibration (corgidrp.data.CoreThroughputCalibration): core throughput calibration object 
            containing off-axis PSFs.
        klip_params (dict): dictionary containing the same KLIP parameters used for PSF subtraction. Must 
            contain the keywords 'mode','annuli','subsections','movement','numbasis','outdir'.
            See corgidrp.l3_to_l4.do_psf_subtraction() for descriptions of each of these parameters.
        inject_snr (int, optional): SNR at which to inject fake PSFs. Defaults to 5.
        sep_spacing (float, optional): multiples of the FWHM at which to space separation samples. Defaults to 3. 
            Overridden by passing in explicit separations to the seps keyword.
        n_pas (int,optional): number of evenly spaced position angles at which to inject PSFs. Defaults to 5. 
            Overridden by in explicit PAs to the pas keyword.
        seps (np.array, optional): Separations (in pixels from the star center) at which to inject fake 
            PSFs. If not provided, a linear spacing of separations between the IWA & OWA will be chosen.
        pas (np.array, optional): Position angles (in degrees counterclockwise from north/up) at which to inject fake 
            PSFs at each separation. Defaults to [0.,90.,180.,270.].
        cand_locs (list of tuples, optional): Locations of known off-axis sources, so we don't inject a fake 
            PSF too close to them. This is a list of tuples (sep_pix,pa_degrees) for each source. Defaults to [].
        num_processes (int): number of processes for parallelizing the PSF subtraction
        
    Returns: 
        np.array: array of shape (N,n_seps,2), where N is 1 + the number of KL mode truncation choices and n_seps 
        is the number of separations sampled. Index 0 contains the separations sampled, and each following index
        contains the dimensionless KLIP throughput and FWHM in pixels measured at each separation for each KL mode 
        truncation choice.
    """
    
    if sci_dataset_in[0].ext_hdr['CFAMNAME'] == '1F':
        lam = 573.8e-9 #m
    else:
        raise NotImplementedError("Only band 1 observations using CFAMNAME 1F are currently configured.")

    d = 2.36 #m  
    pixscale_mas = sci_dataset_in[0].ext_hdr['PLTSCALE']
    fwhm_mas = 1.22 * lam / d * 206265. * 1000.
    fwhm_pix = fwhm_mas / pixscale_mas  
    res_elem = sep_spacing * fwhm_pix # pix
    
    if seps is None:
        if sci_dataset_in[0].ext_hdr['LSAMNAME'] == 'NFOV':
            owa_mas = 450. 
            owa_pix = owa_mas / pixscale_mas   
        else:
            raise NotImplementedError("Automatic separation choices only configured for NFOV observations.")
        
        if sci_dataset_in[0].ext_hdr['FPAMNAME'] == 'HLC12_C2R1':
            iwa_mas = 140. 
            iwa_pix = iwa_mas / pixscale_mas 
        else:
            raise NotImplementedError("Automatic separation choices only configured for NFOV observations.")
        
        seps = np.arange(iwa_pix,owa_pix,res_elem) # Some linear spacing between the IWA & OWA, around 5x the fwhm
    if pas is None:
        pas = np.linspace(0.,360.,n_pas+1)[:-1] # Some linear spacing between the IWA & OWA, around 5x the fwhm

    thrupts = []
    outfwhms = []
    for k,klmode in enumerate(klip_params['numbasis']):
        
        sci_dataset = sci_dataset_in.copy()

        rolls = [frame.pri_hdr['ROLL'] for frame in sci_dataset]
        
        # Measure noise at each separation in psf subtracted dataset (for this kl mode)
        noise_vals = measure_noise(psfsub_dataset[0],seps,fwhm_pix,k,cand_locs)
        
        # Inject planets:
        seppas_skipped = []

        this_klmode_seppas = []
        this_klmode_psfmodels = []
        this_klmode_inject_peaks = []
        this_klmode_psfcenxy = []

        for i,frame in enumerate(sci_dataset):
            this_klmode_seppas.append([])
            this_klmode_psfmodels.append([])
            this_klmode_inject_peaks.append([])
            this_klmode_psfcenxy.append([])

            # Initialize PA offset to spiral the injections
            pa_off = 0.
            pa_step = 360. / len(pas) / 3
            
            for s,sep in enumerate(seps):
                
                noise = noise_vals[s]
                inject_peak = noise * inject_snr

                for pa in pas:
                    pa = (pa + pa_off) % 360.
                    inject_loc = (sep,pa)

                    if inject_loc in seppas_skipped:
                        continue
                    
                    # Check that we're not too close to a candidate during the first frame
                    if i==0:
                        too_close = False
                        for cand_sep, cand_pa in cand_locs:
                            # Account for telescope roll angles, skip if any are too close
                            for roll in rolls:
                                cand_pa_adj = cand_pa
                                dist = get_polar_dist((cand_sep,cand_pa_adj),inject_loc)
                                if dist < res_elem:
                                    too_close=True
                                    break
                            if too_close:
                                break
                        if too_close:
                            seppas_skipped.append(inject_loc)
                            continue
                    
                    frame, psf_model, psf_cenxy = inject_psf(frame, ct_calibration, 
                                                    inject_peak, *inject_loc)

                    # Save these for later
                    this_klmode_psfmodels[i].append(psf_model.copy())
                    this_klmode_seppas[i].append(inject_loc)
                    this_klmode_inject_peaks[i].append(inject_peak)
                    this_klmode_psfcenxy[i].append(psf_cenxy)
                pa_off += pa_step

            sci_dataset[i].data[:] = frame.data[:]
                    
        # Debugging things
        psfmodels_arr = np.array(this_klmode_psfmodels)
        seppas_arr = np.array(this_klmode_seppas)
        inj_peaks = np.array(this_klmode_inject_peaks)
        psfcenxys = np.array(this_klmode_psfcenxy)
        psfmodel_sums = np.sum(psfmodels_arr,axis=(2,3))
        psfmodel_peaks = np.max(psfmodels_arr,axis=(2,3))

        # Init pyklip dataset
        pyklip_dataset = PyKLIPDataset(sci_dataset,psflib_dataset=ref_dataset_in)
        
        # Run pyklip
        klip_dataset(pyklip_dataset, outputdir=klip_params['outdir'],
                                annuli=klip_params['annuli'], subsections=klip_params['subsections'], 
                                movement=klip_params['movement'], 
                                numbasis=[klmode],
                                calibrate_flux=False, mode=klip_params['mode'],
                                psf_library=pyklip_dataset._psflib,
                                fileprefix=f"FAKE_{klmode}KLMODES",
                                numthreads=num_processes)
        
        # Get photometry of each injected source

        # Load pyklip output
        pyklip_fpath = os.path.join(klip_params['outdir'],f"FAKE_{klmode}KLMODES-KLmodes-all.fits")
        pyklip_data = fits.getdata(pyklip_fpath)[0]
        pyklip_hdr = fits.getheader(pyklip_fpath)

        # Measure background via sigma clip 
        n_loops = 5
        masked_data = pyklip_data.copy()
        for n in range(n_loops):
            std = np.nanstd(masked_data)
            med = np.nanmedian(masked_data)
            clip_thresh = 3 * std
            masked_data = np.where(np.abs(masked_data-med)>clip_thresh,np.nan,masked_data)
        
        # Subtract median
        bg_level = np.nanmedian(masked_data)
        medsubtracted_data = pyklip_data - bg_level
        
        # # Plot Psf subtraction with fakes
        # if klip_params['mode'] == 'RDI':
        #     analytical_result = rotate(sci_dataset[0].data - ref_dataset_in[0].data,-rolls[0],reshape=False,cval=np.nan)
        # elif klip_params['mode'] == 'ADI':
        #     analytical_result = shift((rotate(sci_dataset[0].data - sci_dataset[1].data,-rolls[0],reshape=False,cval=0) + rotate(sci_dataset[1].data - sci_dataset[0].data,-rolls[1],reshape=False,cval=0)) / 2,
        #                     [0.5,0.5],
        #                     cval=np.nan)
        # elif klip_params['mode'] == 'ADI+RDI':
        #     analytical_result = (rotate(sci_dataset[0].data - (sci_dataset[1].data/2+ref_dataset_in[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(sci_dataset[1].data - (sci_dataset[0].data/2+ref_dataset_in[0].data/2),-rolls[1],reshape=False,cval=0)) / 2

        # import matplotlib.pyplot as plt
        # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
        # im0 = axes[0].imshow(medsubtracted_data,origin='lower')
        # plt.colorbar(im0,ax=axes[0],shrink=0.8)
        # locsxy = seppa2xy(seppas_arr[0,:,0],seppas_arr[0,:,1],pyklip_hdr['PSFCENTX'],pyklip_hdr['PSFCENTY'])
        # axes[0].scatter(locsxy[0],locsxy[1],label='Injected PSFs',s=1,c='r',marker='x')
        # axes[0].set_title(f'Output data')
        # axes[0].legend()
        # im1 = axes[1].imshow(analytical_result,origin='lower')
        # plt.colorbar(im1,ax=axes[1],shrink=0.8)
        # axes[1].set_title('Analytical result')
        # im2 = axes[2].imshow(medsubtracted_data - analytical_result,origin='lower')
        # plt.colorbar(im2,ax=axes[2],shrink=0.8)
        # axes[2].set_title('Difference')
        # plt.suptitle(f'PSF Subtraction {klip_params["mode"]} ({klmode} KL Modes)')
        # plt.show()
        
        # After psf subtraction
        # this_klmode_peakin = []
        # this_klmode_peakout = []
        # this_klmode_sumin = []
        # this_klmode_influxs = []
        # this_klmode_outfluxs = []
        this_klmode_infwhms = []
        this_klmode_outfwhms = []
        this_klmode_thrupts = []
        for ll,loc in enumerate(this_klmode_seppas[0]):
            
            psf_model = this_klmode_psfmodels[0][ll]
            
            # Pad psf model with zeros so we can measure background
            model_shape = np.array(psf_model.shape)
            cutout_shape = model_shape * 2 + 1
            cutoutcenyx = cutout_shape/2. - 0.5

            psf_model_padded = np.zeros(cutout_shape)
            start_ind_model = (cutoutcenyx-model_shape//2).astype(int)
            end_ind_model = (start_ind_model + model_shape).astype(int)
            x1_model,y1_model = start_ind_model
            x2_model,y2_model = end_ind_model

            psf_model_padded[y1_model:y2_model,
                             x1_model:x2_model] = psf_model

            # Crop data around location to be same as psf_model cutout
            locxy = seppa2xy(*loc,pyklip_hdr['PSFCENTX'],pyklip_hdr['PSFCENTY'])

            # Crop the data, pad with nans if we're cropping over the edge
            cutout = np.zeros_like(psf_model_padded)
            cutout[:] = np.nan
            cutout_starty, cutout_startx = (0,0)
            cutout_endy, cutout_endx = cutout.shape

            data_shape = medsubtracted_data.shape
            data_center_indyx = np.array([locxy[1],locxy[0]]).astype(int)
            data_start_indyx = (data_center_indyx - cutout_shape//2)
            data_end_indyx = (data_start_indyx + cutout_shape)
            data_starty,data_startx = data_start_indyx
            data_endy,data_endx = data_end_indyx
            
            if data_starty < 0:
                cutout_starty = -data_starty
                data_starty = 0
            
            if data_startx < 0:
                cutout_startx = -data_startx
                data_startx = 0
            
            if data_endy >= data_shape[0]:
                y_overhang = data_endy - medsubtracted_data.shape[0]
                cutout_endy = cutout_shape[0] - y_overhang
                data_endy = data_shape[0]

            if data_endx >= data_shape[1]:
                x_overhang = data_endx - medsubtracted_data.shape[1]
                cutout_endx = cutout_shape[1] - x_overhang
                data_endx = data_shape[1]


            cutout[cutout_starty:cutout_endy,
                        cutout_startx:cutout_endx] = medsubtracted_data[data_starty:data_endy,
                                                            data_startx:data_endx]
            
            # import matplotlib.pyplot as plt
            # fig,ax = plt.subplots(1,2,
            #                       sharey=True,
            #                       layout='constrained',
            #                       figsize=(8,4))
            # im0 = ax[0].imshow(psf_model_padded,origin='lower')
            # plt.colorbar(im0,ax=ax[0])
            # ax[0].set_title('PSF Model')
            # im1 = ax[1].imshow(cutout,origin='lower')
            # plt.colorbar(im1,ax=ax[1])
            # ax[1].set_title('Data Cutout')
            # plt.show()
            
            # Using pyklip.fakes.gaussfit2d
            preklip_peak, pre_fwhm, pre_xfit, pre_yfit = gaussfit2d(
                psf_model_padded, 
                cutoutcenyx[1], 
                cutoutcenyx[0], 
                searchrad=5, 
                guessfwhm=fwhm_pix, 
                guesspeak=inject_peak, 
                refinefit=True)

            postklip_peak, post_fwhm, post_xfit, post_yfit = gaussfit2d(
                cutout, 
                cutoutcenyx[1], 
                cutoutcenyx[0], 
                searchrad=5, 
                guessfwhm=fwhm_pix, 
                guesspeak=inject_peak, 
                refinefit=True) 

            # Get total counts from 2D gaussian fit
            preklip_counts = np.pi * preklip_peak * pre_fwhm**2 / 4. / np.log(2.)
            postklip_counts = np.pi * postklip_peak * post_fwhm**2 / 4. / np.log(2.)
            
            thrupt = postklip_counts/preklip_counts

            # this_klmode_peakin.append(preklip_peak)
            # this_klmode_peakout.append(postklip_peak)
            # this_klmode_sumin.append(np.sum(psf_model))
            # this_klmode_influxs.append(preklip_counts)
            # this_klmode_outfluxs.append(postklip_counts)
            this_klmode_infwhms.append(pre_fwhm)
            this_klmode_outfwhms.append(post_fwhm)
            this_klmode_thrupts.append(thrupt)

        seppas_arr = np.array(this_klmode_seppas[0])
        seps_arr = seppas_arr[:,0]

        # # Plot injected and recovered peaks
        # import matplotlib.pyplot as plt
        # fig,ax = plt.subplots()
        # plt.scatter(seps_arr,this_klmode_influxs,label='Injected counts')
        # plt.scatter(seps_arr,this_klmode_outfluxs,label='Recovered counts')
        # plt.xlabel('separation (pixels)')
        # plt.legend()
        # plt.show()

        # # Plot injected and recovered peaks
        # fig,ax = plt.subplots()
        # plt.scatter(seps_arr,this_klmode_peakin,label='Injected peaks')
        # plt.scatter(seps_arr,this_klmode_peakout,label='Recovered peaks')
        # plt.xlabel('separation (pixels)')
        # plt.legend()
        # plt.show()

        mean_thrupts = []
        mean_outfwhms = []
        # TODO: If no measurements available for a given sep
        # show warning and add np.nan 
        for sep in np.unique(seps_arr):
            this_sep_thrupts = np.where(seps_arr==sep,this_klmode_thrupts,np.nan)
            mean_thrupt = np.nanmedian(this_sep_thrupts)
            mean_thrupts.append(mean_thrupt)

            this_sep_outfwhms = np.where(seps_arr==sep,this_klmode_outfwhms,np.nan)
            mean_outfwhm = np.nanmedian(this_sep_outfwhms)
            mean_outfwhms.append(mean_outfwhm)

        thrupts.append(mean_thrupts)
        outfwhms.append(mean_outfwhms)

    thrupt_arr = np.array([[(sep,sep) for sep in seps],*[[(thrupts[kk][ss],outfwhms[kk][ss]) for ss in range(len(seps))] for kk in range(len(klip_params['numbasis']))]])

    return thrupt_arr



import os
import pytest
import corgidrp
import numpy as np
import corgidrp.detector as detector
from astropy.time import Time
from corgidrp.l2a_to_l2b import desmear
from corgidrp.mocks import create_default_L1_headers
from corgidrp.data import Image, Dataset, DetectorParams

old_err_tracking = corgidrp.track_individual_errors

def test_desmear():
    # Tolerance for comparisons
    tol = 1e-12

    corgidrp.track_individual_errors = False

    print("Testing desmear step function")

    detector_params = DetectorParams({}, date_valid=Time("2023-11-01 00:00:00"))
    rowreadtime_sec = detector_params.params['ROWREADT']

    #make a flux map
    size = 1024
    background_flux = 10
    foreground_flux = 100
    xx, yy = np.mgrid[:size, :size]
    circle = ((xx - size//2)**2 + (yy - size//2)**2 ) < (size//4)**2
    flux = background_flux * np.ones([size,size]) + (foreground_flux - background_flux) * circle

    #make a truth frame
    err = np.ones([1024,1024]) *0.5
    dq = np.zeros([1024,1024], dtype = np.uint16)
    prhd, exthd = create_default_L1_headers()
    e_t=exthd['EXPTIME']
    unsmeared_frame = e_t*flux

    #simulate the smearing
    smear = np.zeros_like(flux)
    m = len(smear)
    for r in range(m):
        columnsum = 0
        for i in range(r+1):
            columnsum = columnsum + rowreadtime_sec*flux[i,:] 
        smear[r,:] = columnsum
    smeared_frame = unsmeared_frame + smear

    image1 = Image(smeared_frame, pri_hdr = prhd, ext_hdr = exthd, err = err,         dq = dq)
    dataset_smeared = Dataset([image1])
    
    assert type(dataset_smeared) == corgidrp.data.Dataset

    # check the header keyword hasn't been toggled yet
    for frame in dataset_smeared:
        assert not frame.ext_hdr['DESMEAR']

    # Apply desmear correction
    dataset_desmear = desmear(dataset_smeared, detector_params)

    assert type(dataset_desmear) == corgidrp.data.Dataset

    assert(np.max(np.abs(dataset_desmear.all_data[0] - unsmeared_frame)) < tol)

    # check the header keyword is toggled
    for frame in dataset_desmear:
        assert frame.ext_hdr['DESMEAR']

if __name__ == '__main__':
    test_desmear()
    



import numpy as np
import astropy.wcs as wcs

# A file that holds the functions that transmogrify l2b data to l3 data 
import numpy as np

def create_wcs(input_dataset, astrom_calibration):
    """
    
    Create the WCS headers for the dataset.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2b-level)
        astrom_calibration (corgidrp.data.AstrometricCalibration): an astrometric calibration file for the input dataset

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the WCS headers added
    """
    updated_dataset = input_dataset.copy()

    northangle = astrom_calibration.northangle
    platescale_x, platescale_y = astrom_calibration.platescale
    ra_offset, dec_offset = astrom_calibration.avg_offset

    # create wcs for each image in the dataset
    for image in updated_dataset:

        im_data = image.data
        image_y, image_x = im_data.shape
        center_pixel = [(image_x-1) // 2, (image_y-1) // 2]
        target_ra, target_dec = image.pri_hdr['RA'], image.pri_hdr['DEC']
        corrected_ra, corrected_dec = target_ra - ra_offset, target_dec - dec_offset
        roll_ang = image.pri_hdr['ROLL']

        vert_ang = np.radians(northangle + roll_ang)  ## might be -roll_ang
        pc = np.array([[-np.cos(vert_ang), np.sin(vert_ang)], [np.sin(vert_ang), np.cos(vert_ang)]])
        cdmatrix_x = pc * (platescale_x * 0.001) / 3600.
        cdmatrix_y = pc * (platescale_y * 0.001) / 3600.


        # create dictionary with wcs information
        wcs_info = {}
        wcs_info['CD1_1'] = cdmatrix_x[0,0]
        wcs_info['CD1_2'] = cdmatrix_x[0,1]
        wcs_info['CD2_1'] = cdmatrix_y[1,0]
        wcs_info['CD2_2'] = cdmatrix_y[1,1]

        wcs_info['CRPIX1'] = center_pixel[0]
        wcs_info['CRPIX2'] = center_pixel[1]

        wcs_info['CTYPE1'] = 'RA---TAN'
        wcs_info['CTYPE2'] = 'DEC--TAN'

        wcs_info['CDELT1'] = (platescale_x * 0.001) / 3600  ## converting to degrees
        wcs_info['CDELT2'] = (platescale_y * 0.001) / 3600

        wcs_info['CRVAL1'] = corrected_ra
        wcs_info['CRVAL2'] = corrected_dec

        wcs_info['PLTSCALE'] = np.mean([platescale_x, platescale_y])  ## [mas] / pixel

        # update the image header with wcs information
        for key, value in wcs_info.items():
            image.ext_hdr[key] = value

    # update the dataset with new header entries an dhistory message
    history_msg = 'WCS created'

    updated_dataset.update_after_processing_step(history_msg)

    return updated_dataset

def divide_by_exptime(input_dataset):
    """
    
    Divide the data by the exposure time to get the units in electrons/s

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2b-level)

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the data in units of electrons/s
    """
    data = input_dataset.copy()

    all_data_new = np.zeros(data.all_data.shape)
    all_err_new = np.zeros(data.all_err.shape)

    for i in range(len(data.frames)):
        exposure_time = float(data.frames[i].ext_hdr['EXPTIME'])

        data.frames[i].data = data.frames[i].data / exposure_time
        # data.frames[i].err = data.frames[i].err / exposure_time
        scale_factor = (1/exposure_time) * np.ones([data.frames[i].err.shape[1], data.frames[i].err.shape[2]])
        # print(data.frames[i].err.shape)
        # print(data.frames[i].err[0])
        # print(scale_factor.shape)
        #data.frames[i].rescale_error(data.frames[i].err, 'normalized by the exposure time')
        data.frames[i].rescale_error(scale_factor, 'normalized by the exposure time')

        all_data_new[i] = data.frames[i].data
        all_err_new[i] = data.frames[i].err

        data.frames[i].ext_hdr.set('BUNIT', 'photoelectrons/s')
    
    history_msg = 'divided by the exposure time'
    data.update_after_processing_step(history_msg, new_all_data = all_data_new, new_all_err = all_err_new)


    return data

def update_to_l3(input_dataset):
    """
    Updates the data level to L3. Only works on L2b data.

    Currently only checks that data is at the L2b level

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2b-level)

    Returns:
        corgidrp.data.Dataset: same dataset now at L3-level
    """
    # check that we are running this on L1 data
    for orig_frame in input_dataset:
        if orig_frame.ext_hdr['DATALVL'] != "L2b":
            err_msg = "{0} needs to be L2b data, but it is {1} data instead".format(orig_frame.filename, orig_frame.ext_hdr['DATALVL'])
            raise ValueError(err_msg)

    # we aren't altering the data
    updated_dataset = input_dataset.copy(copy_data=False)

    for frame in updated_dataset:
        # update header
        frame.ext_hdr['DATALVL'] = "L3"
        # update filename convention. The file convention should be
        # "CGI_[dataleel_*]" so we should be same just replacing the just instance of L1
        frame.filename = frame.filename.replace("_L2b", "_L3_", 1)

    history_msg = "Updated Data Level to L3"
    updated_dataset.update_after_processing_step(history_msg)

    return updated_dataset


# This module is written to do an absolute flux calibration observing a standard star having CALSPEC data.
import glob
import os
import numpy as np
from astropy.io import fits, ascii
from astropy import wcs
from astropy.io import fits, ascii
from astropy.coordinates import SkyCoord
import corgidrp
from photutils.aperture import CircularAperture
from photutils.background import LocalBackground
from photutils.psf import fit_2dgaussian
from scipy import integrate
from corgidrp.astrom import centroid_with_roi
import urllib

# Dictionary of anticipated bright and dim CASLPEC standard star names and corresponding fits names
calspec_names= {
# bright standards
'109 Vir': '109vir_stis_005.fits',
'Vega': 'alpha_lyr_stis_011.fits',
'Eta Uma': 'etauma_stis_008.fits',
'Lam Lep': 'lamlep_stis_008.fits',
'KSI2 CETI': 'ksi2ceti_stis_008.fits',
# dim standards
'TYC 4433-1800-1': '1808347_stiswfc_006.fits',
'TYC 4205-1677-1': '1812095_stisnic_008.fits',
'TYC 4212-455-1': '1757132_stiswfc_006.fits',
'TYC 4209-1396-1': '1805292_stisnic_008.fits',
'TYC 4413-304-1': 'p041c_stisnic_010.fits',
'UCAC3 313-62260': 'kf08t3_stisnic_005.fits',
'BPS BS 17447-0067': '1802271_stiswfcnic_006.fits',
'TYC 4424-1286-1': '1732526_stisnic_009.fits',
'GSC 02581-02323': 'p330e_stiswfcnic_007.fits',
'TYC 4207-219-1': '1740346_stisnic_005.fits'
}

calspec_url = 'https://archive.stsci.edu/hlsps/reference-atlases/cdbs/current_calspec/'

def get_calspec_file(star_name):
    """
    download the corresponding CALSPEC fits file and return the file path
    
    Args:
        star_name (str): 
    
    Returns:
        str: file path
    """
    if star_name not in calspec_names:
        raise ValueError('{0} is not in list of anticipated standard stars {1}, please check naming'.format(star_name, calspec_names.keys()) )
    fits_name = calspec_names.get(star_name)
    # TODO: be flexible with the version of the calspec fits file, so essentially, the number in the name should not matter
    fits_url = calspec_url + fits_name
    try:
        calspec_dir = os.path.join(os.path.dirname(corgidrp.config_filepath), "calspec_data")
        if not os.path.exists(calspec_dir):
            os.mkdir(calspec_dir)
        file_name, headers = urllib.request.urlretrieve(fits_url, filename =  os.path.join(calspec_dir, fits_name))
    except:
        raise Exception("cannot access CALSPEC archive web page and/or download {0}".format(fits_name))
    return file_name

def get_filter_name(image):
    """
    return the name of the transmission curve csv file of used color filter
    
    Args:
        image (corgidrp.image): image of the observed calstar
    
    Returns:
        str: filepath of the selected filter curve
    """
    datadir = os.path.join(os.path.dirname(__file__), "data", "filter_curves")
    filters = os.path.join(datadir, "*.csv")
    filter = image.ext_hdr['CFAMNAME']
    filter_names = os.listdir(datadir)

    filter_name = [name for name in filter_names if filter in name]
    if filter_name == []:
        raise ValueError("there is no filter available with name {0}".format(filter))
    else:
        return os.path.join(datadir,filter_name[0])

def read_filter_curve(filter_filename):
    """
    read the transmission curve csv file of the color filters
    
    Args:
        filter_filename (str): file name of the transmission curve data
    
    Returns:
        lambda_nm (np.array): wavelength in unit Angstroem
        transmission (np.array): transmission of the filter < 1
    """
    tab = ascii.read(filter_filename, format='csv', header_start = 3, data_start = 4)
    lambda_nm = tab['lambda_nm'].data #unit nm
    transmission = tab['%T'].data
    return lambda_nm * 10 , transmission/100.

def read_cal_spec(calspec_filename, filter_wavelength):
    """
    read the calspec flux density data interpolated on the wavelength grid of the transmission curve
    
    Args:
        calspec_filename (str): file name of the CALSPEC fits file
        filter_wavelength (np.array): wavelength grid of the transmission curve in unit Angstroem
    
    Returns:
        np.array: flux density in Jy interpolated on the wavelength grid of the transmission curve 
        in CALSPEC units erg/(s * cm^2 * AA)
    """
    hdulist = fits.open(calspec_filename)
    data = hdulist[1].data
    hdulist.close()
    w = data['WAVELENGTH'] #wavelength in Angstroem
    flux = data['FLUX']
    flux = flux[(w<=filter_wavelength[-1]) & (w>=filter_wavelength[0])] #erg/(s*cm^2*AA)
    w = w[(w<=filter_wavelength[-1]) & (w>=filter_wavelength[0])]

    #interpolate on transmission curve wavelengths
    flux_inter = np.interp(filter_wavelength, w, flux)
    
    return flux_inter

def calculate_band_flux(filter_curve, calspec_flux, filter_wavelength):
    """
    calculate the average band flux of a calspec source in the filter band, see convention A in Gordon et al. (2022)
    TBC if needed at all
    
    Args:
        filter_curve (np.array): filter transmission curve over the filter_wavelength
        calspec_flux (np.array): converted flux in units of erg/(s*cm^2*AA) of the calpec source in the filter band
        filter_wavelength (np.array): wavelengths in units Angstroem in the filter band 
    
    Returns:
        float: average band flux of the calspec star in unit erg/(s*cm^2*AA)
    """
    multi_flux = calspec_flux * filter_curve * filter_wavelength
    multi_band = filter_curve * filter_wavelength
    aver_flux = integrate.simpson(multi_flux, x=filter_wavelength)/integrate.simpson(multi_band, x=filter_wavelength)
    
    return aver_flux

def calculate_effective_lambda(filter_curve, calspec_flux, filter_wavelength):
    """
    calculate the effective wavelength of a calspec source in the filter band, see convention A in Gordon et al. (2022)
    TBC if needed at all
    
    Args:
        filter_curve (np.array): filter transmission curve over the filter_wavelength
        calspec_flux (np.array): converted flux in units of the calpec source in the filter band
        filter_wavelength (np.array): wavelengths in units nm in the filter band 
    
    Returns:
        float: effective wavelength in unit Angstroem
    """
    multi_flux = calspec_flux * filter_curve * np.square(filter_wavelength)
    multi_band = calspec_flux * filter_curve * filter_wavelength
    eff_lambda = integrate.simpson(multi_flux, x=filter_wavelength)/integrate.simpson(multi_band, x=filter_wavelength)
    
    return eff_lambda

def calculate_pivot_lambda(filter_curve, filter_wavelength):
    """
    calculate the reference pivot wavelength of the filter band, see convention B in Gordon et al. (2022)
    
    Args:
        filter_curve (np.array): filter transmission curve over the filter_wavelength
        filter_wavelength (np.array): wavelengths in unit Angstroem in the filter band 
    
     Returns:
        float: pivot wavelength in unit Angstroem
    """
    multi_flux = filter_curve * filter_wavelength
    multi_band = filter_curve / filter_wavelength
    piv_lambda = np.sqrt(integrate.simpson(multi_flux, x=filter_wavelength)/integrate.simpson(multi_band, x=filter_wavelength))
    
    return piv_lambda

def calculate_flux_ref(filter_wavelength, calspec_flux, wave_ref):
    """
    calculate the flux at the reference wavelength of the filter band
    
    Args:
        filter_wavelength (np.array): wavelengths in unit Angstroem in the filter band 
        calspec_flux (np.array): converted flux in units of the calpec source in the filter band
        wave_ref (float): reference wavelength in unit Angstroem
    
    Returns:
        float: flux at reference wavelength in unit erg/(s*cm^2*AA)
    """
    
    flux_ref = np.interp(wave_ref, filter_wavelength, calspec_flux)
    return flux_ref

def calculate_vega_mag(source_flux, filter_file):
    """
    determine the apparent Vega magnitude of the source with known flux in CALSPEC units (erg/(s * cm^2 *AA)
    in the used filter band.
    
    Args:
        source_flux (float): source flux usually determined by applying the FluxcalFactor
        filter_file (str): name of the file with the transmission data of the corresponding color filter
    
    Returns:
        float: the apparent VEGA magnitude
    """
    
    wave, filter_trans = read_filter_curve(filter_file)
    # calculate the flux of VEGA and the source star from the user given CALSPEC file binned on the wavelength grid of the filter
    vega_filepath = get_calspec_file('Vega')
    vega_sed = read_cal_spec(vega_filepath, wave)

    vega_flux = calculate_band_flux(filter_trans, vega_sed, wave)
    #calculate apparent vega magnitude
    vega_mag = -2.5 * np.log10(source_flux/vega_flux)
    
    return vega_mag


def compute_color_cor(filter_curve, filter_wavelength , flux_ref, wave_ref, flux_source):
    """
    Compute the color correction factor K given the filter bandpass, reference spectrum (CALSPEC),
    and source spectrum model.  To use this color correction, divide the flux density
    for a band by K.  Such color corrections are needed to compute the correct
    flux density at the reference wavelength for a source with the flux_source
    spectral shape in the photometric convention that provides the flux density
    at a reference wavelength (convention B, see Gordon et al. 2022, The Astronomical Journal 163:267, for details).
    Thus the flux density value found by applying the calibration factor on the found detected electrons 
    of an arbitrary source should be divided by K (for the appropriate filter and spectral shape) 
    to produce the flux density at the reference wavelength of the filter. 
    The color correction adjusts the calibration factor to align the reference spectral shape 
    with the current source, which results in the correct flux density at the reference wavelength.

    Args:
    filter_curve (np.array): transmission of the filter bandpass
    filter_wavelength (np.array): the wavelengths of the filter bandpass, flux_ref, and flux_source in unit Angstroem
    flux_ref (np.array): reference flux density F(lambda) as a function of wavelength
    wave_ref (float): reference wavelength in unit Angstroem
    flux_source (np.array): source flux density F(lambda) as a function of wavelength in CALSPEC unit erg/(s * cm^2 * AA)
    
    Returns:
        float: color correction factor K
    """
    # get the flux densities at the reference wavelength
    flux_source_lambda_ref = calculate_flux_ref(filter_wavelength, flux_source, wave_ref)
    flux_ref_lambda_ref = calculate_flux_ref(filter_wavelength, flux_ref, wave_ref)

    # compute the top and bottom integrals
    int_source = integrate.simpson(filter_wavelength * filter_curve * flux_source / flux_source_lambda_ref, x=filter_wavelength)
    int_ref = integrate.simpson(filter_wavelength * filter_curve * flux_ref / flux_ref_lambda_ref, x=filter_wavelength)

    return int_source / int_ref

def calculate_band_irradiance(filter_curve, calspec_flux, filter_wavelength):
    """
    calculate the integrated band flux, irradiance of a calspec source in the filter band
    to determine the apparent magnitude
    
    Args:
        filter_curve (np.array): filter transmission curve over the filter_wavelength
        calspec_flux (np.array): converted flux in units of erg/(s*cm^2*AA) of the calpec source in the filter band
        filter_wavelength (np.array): wavelengths in units Angstroem in the filter band 
    
    Returns:
        float: band irradiance of the calspec star in unit erg/(s*cm^2)
    """
    multi_flux = calspec_flux * filter_curve
    irrad = integrate.simpson(multi_flux, x=filter_wavelength)
    
    return irrad


def aper_phot(image, encircled_radius, frac_enc_energy=1., method='subpixel', subpixels=5,
              background_sub=False, r_in=5, r_out=10, centering_method='xy', centroid_roi_radius=5,
              centering_initial_guess=None):
    """
    Returns the flux in photo-electrons of a point source, either by placing an aperture using a 
        centroiding method, or by using WCS information.
    Background subtraction can be done optionally using a user defined circular annulus.
    
    Parameters:
        image (corgidrp.data.Image): combined source exposure image
        encircled_radius (float): pixel radius of the circular aperture to sum the flux
        frac_enc_energy (float): fraction of encircled energy inside the encircled_radius of the PSF, inverse aperture correction, 0...1
        method (str): {‘exact’, ‘center’, ‘subpixel’}, The method used to determine the overlap of the aperture on the pixel grid, 
        default is 'exact'. For detailed description see https://photutils.readthedocs.io/en/stable/api/photutils.aperture.CircularAnnulus.html
        subpixels (int): For the 'subpixel' method, resample pixels by this factor in each dimension. That is, each pixel is divided 
                         into subpixels**2 subpixels. This keyword is ignored unless method='subpixel', default is 5
        background_sub (boolean): background can be determine from a circular annulus (default: False).
        r_in (float): inner radius of circular annulus in pixels, (default: 5)
        r_out (float): outer radius of circular annulus in pixels, (default: 10)
        centering_method (str): 'xy' for centroiding or 'wcs' for WCS-based centering.
        centroid_roi_radius (int or float): Half-size of the box around the peak,
                                   in pixels. Adjust based on desired λ/D.
        centering_initial_guess (tuple): (Optional) (x,y) initial guess to perform centroiding.  
    
    Returns:
        tuple: (flux, flux_err) or (flux, flux_err, back) if background_sub is True.
    """
    if frac_enc_energy <= 0 or frac_enc_energy > 1:
        raise ValueError("frac_enc_energy {0} should be within 0 < fee <= 1".format(str(frac_enc_energy)))
    
    # Work on a copy so that background subtraction does not alter original data.
    dat = image.data.copy()
    
    # Determine the center position using either WCS or centroid method.
    if centering_method == 'wcs':
        ra = image.pri_hdr['RA']
        dec = image.pri_hdr['DEC']
        target_skycoord = SkyCoord(ra=ra, dec=dec, unit='deg')
        w = wcs.WCS(image.ext_hdr)
        pos = wcs.utils.skycoord_to_pixel(target_skycoord, w, origin=1)
    elif centering_method == 'xy':
        x_center, y_center = centroid_with_roi(image.data, centroid_roi_radius, centering_initial_guess)
        pos = (x_center, y_center)
    else:
        raise ValueError("Invalid centering_method. Choose 'xy' or 'wcs'.")
    
    # Optionally subtract the background.
    if background_sub:
        #This is essentially the median in a circular annulus 
        bkg = LocalBackground(r_in, r_out)
        back = bkg(dat, pos[0], pos[1], mask=image.dq.astype(bool))
        dat -= back

    # Create the circular aperture and compute photometry.
    aper = CircularAperture(pos, encircled_radius)
    aperture_sums, aperture_sums_errs = aper.do_photometry(
        dat,
        error=image.err[0],
        mask=image.dq.astype(bool),
        method=method,
        subpixels=subpixels
    )
    
    flux = aperture_sums[0] / frac_enc_energy
    flux_err = aperture_sums_errs[0] / frac_enc_energy
    
    if background_sub:
        return flux, flux_err, back
    else:
        return flux, flux_err


def phot_by_gauss2d_fit(image, fwhm, fit_shape=None, background_sub=False, r_in=5,
                        r_out=10, centering_method='xy', centroid_roi_radius=5,
                        centering_initial_guess=None):
    """
    Returns the flux in photo-electrons using a 2D Gaussian fit. Finds the star
        center either by placing an aperture using a centroiding method, or by using 
        WCS information.
    Allows optional background subtraction and selection of centering method.
    
    Parameters:
        image (corgidrp.data.Image): the source image.
        fwhm (float): Estimated full-width at half maximum.
        fit_shape (int or tuple, optional): Fitting region shape. If a scalar, then 
            it is assumed to be a square. If `None`, then the shape of the input 'data'.
            It must be an odd value and should be much bigger than fwhm.
        background_sub (bool): If True, subtract background.
        r_in (float): Inner annulus radius.
        r_out (float): Outer annulus radius.
        centering_method (str): 'xy' or 'wcs' centering.
        centroid_roi_radius (int or float): Half-size of the box around the peak,
                                   in pixels. Adjust based on desired λ/D.
        centering_initial_guess (tuple): (Optional) (x,y) initial guess to perform centroiding.
    
    Returns:
        tuple: (flux, flux_err)
    """
    # Determine the star center via WCS or centroid method.
    if centering_method == 'wcs':
        ra = image.pri_hdr['RA']
        dec = image.pri_hdr['DEC']
        target_skycoord = SkyCoord(ra=ra, dec=dec, unit='deg')
        w = wcs.WCS(image.ext_hdr)
        pos = wcs.utils.skycoord_to_pixel(target_skycoord, w, origin=1)
    elif centering_method == 'xy':
        x_center, y_center = centroid_with_roi(image.data, centroid_roi_radius, centering_initial_guess)
        pos = (x_center, y_center)
    else:
        raise ValueError("Invalid centering_method. Choose 'xy' or 'wcs'.")

    # Work on a copy so that background subtraction does not alter original data.
    dat = image.data.copy()

    if background_sub:
        bkg = LocalBackground(r_in, r_out)
        back = bkg(dat, pos[0], pos[1], mask=image.dq.astype(bool))
        dat -= back

    # fit_2dgaussian: error weighting raises exception if error is zero
    err = image.err[0].copy()
    err[err == 0] = np.finfo(np.float32).eps

    if fit_shape is None:
        fit_shape = image.data.shape[0] - 1

    psf_phot = fit_2dgaussian(dat, xypos=pos, fwhm=fwhm, fit_shape=fit_shape,
                              mask=image.dq.astype(bool), error=err)
    flux = psf_phot.results['flux_fit'][0]
    flux_err = psf_phot.results['flux_err'][0]

    if background_sub:
        return [flux, flux_err, back]
    else:
        return [flux, flux_err]


def calibrate_fluxcal_aper(dataset_or_image, flux_or_irr = 'flux', phot_kwargs=None):
    """
    fills the FluxcalFactors calibration product values for one filter band,
    calculates the flux calibration factors by aperture photometry.
    The band flux values are divided by the found photoelectrons/s.
    Propagates also errors to flux calibration factor calfile.
    Background subtraction can be done optionally using a user defined circular annulus.
    
    The photometry parameters are controlled via the `phot_kwargs` dictionary.
    Defaults are provided below if these parameters are not defined. 
    Accepted keywords:
        'encircled_radius' (float): The radius of the circular aperture used for photometry.
        'frac_enc_energy' (float): The fraction of the total flux expected to be enclosed 
            within the aperture. Must be in the range (0, 1].
        'method' (str): The photometry method to use. For example, 'subpixel' indicates subpixel 
            sampling for the aperture.
        'subpixels' (int): The number of subpixels per pixel to use in the photometry calculation 
            or improved resolution.
        'background_sub' (bool): Flag indicating whether to subtract background using an annulus.
        'r_in' (float): The inner radius of the annulus used for background estimation.
        'r_out' (float): The outer radius of the annulus used for background estimation.
        'centering_method' (str): The method for determining the star's center. Options include 
            'xy' for centroiding or 'wcs' for WCS-based centering.
        'centroid_roi_radius' (int or float): Half-size of the box around the peak,
                                   in pixels. Adjust based on desired λ/D.
        'centering_initial_guess' (tuple): (Optional) (x,y) initial guess to perform centroiding.
    
    Parameters:
        dataset_or_image (corgidrp.data.Dataset or corgidrp.data.Image): Image(s) to compute 
            the calibration factor. Should already be normalized for exposure time.
        flux_or_irr (str, optional): Whether flux ('flux') or in-band irradiance ('irr) should 
            be used.
        phot_kwargs (dict, optional): A dictionary of keyword arguments controlling the aperture 
            photometry function.

    Returns:
        FluxcalFactor (corgidrp.data.FluxcalFactor): A calibration object containing the computed 
            flux calibration factor in (TO DO: what units should this be in)
    """
    if isinstance(dataset_or_image, corgidrp.data.Dataset):
        image = dataset_or_image[0]
        dataset = dataset_or_image
    else:
        image = dataset_or_image
        dataset = corgidrp.data.Dataset([image])
    
    if phot_kwargs is None:
        phot_kwargs = {
            'encircled_radius': 5,
            'frac_enc_energy': 1.0,
            'method': 'subpixel',
            'subpixels': 5,
            'background_sub': False,
            'r_in': 5,
            'r_out': 10,
            'centering_method': 'xy',
            'centroid_roi_radius': 5
        }
    
    star_name = image.pri_hdr["TARGET"]
    filter_name = image.ext_hdr["CFAMNAME"]
    filter_file = get_filter_name(image)
    
    # Read filter and CALSPEC data.
    wave, filter_trans = read_filter_curve(filter_file)
    calspec_filepath = get_calspec_file(star_name) 
    flux_ref = read_cal_spec(calspec_filepath, wave)
    
    if flux_or_irr == 'flux':
        flux = calculate_band_flux(filter_trans, flux_ref, wave)
        image.ext_hdr['BUNIT'] = 'erg/(s * cm^2 * AA)/(electron/s)'
        image.err_hdr['BUNIT'] = 'erg/(s * cm^2 * AA)/(electron/s)'
    elif flux_or_irr == 'irr':
        flux = calculate_band_irradiance(filter_trans, flux_ref, wave)
        image.ext_hdr['BUNIT'] = 'erg/(s * cm^2)/(electron/s)'
        image.err_hdr['BUNIT'] = 'erg/(s * cm^2)/(electron/s)'
    else:
        raise ValueError("Invalid flux method. Choose 'flux' or 'irr'.")
    
    result = aper_phot(image, **phot_kwargs)
    if phot_kwargs.get('background_sub', False):
        ap_sum, ap_sum_err, back = result
    else:
        ap_sum, ap_sum_err = result

    fluxcal_fac = flux / ap_sum
    fluxcal_fac_err = flux / ap_sum**2 * ap_sum_err

    fluxcal_obj = corgidrp.data.FluxcalFactor(
        np.array([[fluxcal_fac]]),
        err=np.array([[[fluxcal_fac_err]]]),
        pri_hdr=image.pri_hdr,
        ext_hdr=image.ext_hdr,
        input_dataset=dataset
    )

    # If background subtraction was performed, set the LOCBACK keyword.
    if phot_kwargs.get('background_sub', False):
        # Here, "back" is the third value returned from phot_by_gauss2d_fit.
        fluxcal_obj.ext_hdr['LOCBACK'] = back

    # Append to or create a HISTORY entry in the header.
    history_entry = "Flux calibration factor was determined by aperture photometry."
    fluxcal_obj.ext_hdr.add_history(history_entry)

    return fluxcal_obj


def calibrate_fluxcal_gauss2d(dataset_or_image, flux_or_irr = 'flux', phot_kwargs=None):
    """
    fills the FluxcalFactors calibration product values for one filter band,
    calculates the flux calibration factors by fitting a 2D Gaussian.
    The band flux values are divided by the found photoelectrons/s.
    Propagates also errors to flux calibration factor calfile.
    Background subtraction can be done optionally using a user defined circular annulus.
    
    All photometry settings are provided via the phot_kwargs dictionary.
    Defaults are provided below if these parameters are not defined. 
    Accepted keywords:
        'fwhm' (float): The expected full width at half maximum.
        'fit_shape' (int or tuple): Fitting region shape.
        'background_sub' (bool): Flag indicating whether to subtract background using an annulus.
        'r_in' (float): The inner radius of the annulus used for background estimation.
        'r_out' (float): The outer radius of the annulus used for background estimation.
        'centering_method' (str): The method for determining the star's center. Options include 
            'xy' for centroiding or 'wcs' for WCS-based centering.
        'centroid_roi_radius' (int or float): Half-size of the box around the peak,
            in pixels. Adjust based on desired λ/D.
        'centering_initial_guess' (tuple): (Optional) (x,y) initial guess to perform centroiding.

    Parameters:
        dataset_or_image (corgidrp.data.Dataset or corgidrp.data.Image): Image(s) to compute 
            the calibration factor. Should already be normalized for exposure time.
        flux_or_irr (str, optional): Whether flux ('flux') or in-band irradiance ('irr) should 
            be used.
        phot_kwargs (dict, optional): A dictionary of keyword arguments controlling the Gaussian 
            photometry function.
                
    Returns:
        FluxcalFactor (corgidrp.data.FluxcalFactor): A calibration object containing the computed 
            flux calibration factor in (TO DO: what units should this be in)
    """
    if isinstance(dataset_or_image, corgidrp.data.Dataset):
        image = dataset_or_image[0]
        dataset = dataset_or_image
    else:
        image = dataset_or_image
        dataset = corgidrp.data.Dataset([image])
    
    if phot_kwargs is None:
        phot_kwargs = {
        'fwhm': 3,
        'fit_shape': None,
        'background_sub': False,
        'r_in': 5,
        'r_out': 10,
        'centering_method': 'xy',
        'centroid_roi_radius': 5
    }

    star_name = image.pri_hdr["TARGET"]
    filter_file = get_filter_name(image)
    
    wave, filter_trans = read_filter_curve(filter_file)
    calspec_filepath = get_calspec_file(star_name)
    flux_ref = read_cal_spec(calspec_filepath, wave)
    
    if flux_or_irr == 'flux':
        flux = calculate_band_flux(filter_trans, flux_ref, wave)
        image.ext_hdr['BUNIT'] = 'erg/(s * cm^2 * AA)/(electron/s)'
        image.err_hdr['BUNIT'] = 'erg/(s * cm^2 * AA)/(electron/s)'
    elif flux_or_irr == 'irr':
        flux = calculate_band_irradiance(filter_trans, flux_ref, wave)
        image.ext_hdr['BUNIT'] = 'erg/(s * cm^2)/(electron/s)'
        image.err_hdr['BUNIT'] = 'erg/(s * cm^2)/(electron/s)'
    else:
        raise ValueError("Invalid flux method. Choose 'flux' or 'irr'.")
    
    if phot_kwargs.get('background_sub', False):
        gauss_sum, gauss_sum_err, back = phot_by_gauss2d_fit(image, **phot_kwargs)
    else:
        gauss_sum, gauss_sum_err = phot_by_gauss2d_fit(image, **phot_kwargs)
    
    fluxcal_fac = flux / gauss_sum
    fluxcal_fac_err = flux / gauss_sum**2 * gauss_sum_err

    fluxcal_obj = corgidrp.data.FluxcalFactor(
        np.array([[fluxcal_fac]]),
        err=np.array([[[fluxcal_fac_err]]]),
        pri_hdr=image.pri_hdr,
        ext_hdr=image.ext_hdr,
        input_dataset=dataset
    )
    
    # If background subtraction was performed, set the LOCBACK keyword.
    if phot_kwargs.get('background_sub', False):
        # Here, "back" is the third value returned from phot_by_gauss2d_fit.
        fluxcal_obj.ext_hdr['LOCBACK'] = back

    # Append to or create a HISTORY entry in the header.
    history_entry = "Flux calibration factor was determined by a Gaussian 2D fit photometry."
    fluxcal_obj.ext_hdr.add_history(history_entry)
    
    return fluxcal_obj



import os
import warnings
import numpy as np
import pandas as pd
from pathlib import Path
from scipy.optimize import curve_fit

from corgidrp import check
import corgidrp.data as data
from corgidrp.data import Image
from corgidrp.detector import slice_section, detector_areas

# Dictionary with constant kgain calibration parameters
kgain_params_default= {
# ROI constants
'rowroi1': 9,
'rowroi2': 1000,
'colroi1': 1199,
'colroi2': 2000,

# read noise bins range limits
'rn_bins1': -200,
'rn_bins2': 201,

# maximum DN value to be included in PTC
'max_DN_val': 13000,

# number of bins in the signal variables
'signal_bins_N': 400,
}

def check_kgain_params(kgain_params):
    """ Checks integrity of kgain parameters in the dictionary kgain_params. 
    
    Args:
        kgain_params (dict):  Dictionary of parameters used for calibrating the k gain.
    """
    
    if 'rowroi1' not in kgain_params:
        raise ValueError('Missing parameter:  rowroi1.')
    if 'rowroi2' not in kgain_params:
        raise ValueError('Missing parameter:  rowroi2.')
    if 'colroi1' not in kgain_params:
        raise ValueError('Missing parameter:  colroi1.')
    if 'colroi2' not in kgain_params:
        raise ValueError('Missing parameter:  colroi2.')
    if 'rn_bins1' not in kgain_params:
        raise ValueError('Missing parameter:  rn_bins1.')
    if 'rn_bins2' not in kgain_params:
        raise ValueError('Missing parameter:  rn_bins2.')
    if 'max_DN_val' not in kgain_params:
        raise ValueError('Missing parameter:  max_DN_val.')
    if 'signal_bins_N' not in kgain_params:
        raise ValueError('Missing parameter:  signal_bins_N.')

    if not isinstance(kgain_params['rowroi1'], (float, int)):
        raise TypeError('rowroi1 is not a number')
    if not isinstance(kgain_params['rowroi2'], (float, int)):
        raise TypeError('rowroi2 is not a number')
    if not isinstance(kgain_params['colroi1'], (float, int)):
        raise TypeError('colroi1 is not a number')
    if not isinstance(kgain_params['colroi2'], (float, int)):
        raise TypeError('colroi2 is not a number')
    if not isinstance(kgain_params['rn_bins1'], (float, int)):
        raise TypeError('rn_bins1 is not a number')
    if not isinstance(kgain_params['rn_bins2'], (float, int)):
        raise TypeError('rn_bins2 is not a number')
    if not isinstance(kgain_params['max_DN_val'], (float, int)):
        raise TypeError('max_DN_val is not a number')
    if not isinstance(kgain_params['signal_bins_N'], (float, int)):
        raise TypeError('signal_bins_N is not a number')

class CalKgainException(Exception):
    """Exception class for calibrate_kgain."""

################### function defs #####################
    
def ptc_bin2(frame_in, mean_frame, binwidth, max_DN):
    """ 
    frame_in is a bias-corrected frame trimmed to the ROI. mean_frame is 
    the scaled high SNR mean frame made from the >=30 frames of uniform 
    exposure time, binwidth is an integer equal to the width of each bin, 
    max_DN is an integer equal to the maximum DN value to be included in 
    the bins.
       
    Args:
      frame_in (np.array): bias corrected frame
      mean_frame (np.array): mean frame of uniform exposure time
      binwidth (int): width of each bin
      max_DN (int): maximum DN value
      
    Returns:
      np.array: mean array
      np.array: mean array
    """
    # calculate the size of output arrays
    rows, cols = frame_in.shape
    out_rows, out_cols = rows // binwidth, cols // binwidth
    
    local_mean_array = np.zeros((out_rows, out_cols))
    local_noise_array = np.zeros((out_rows, out_cols))
        
    # Define the bin edges
    row_bins = np.arange(1, rows + 1, binwidth)
    col_bins = np.arange(1, cols + 1, binwidth)
        
    tot_bins = len(row_bins) * len(col_bins)
    DN_bin = max_DN / tot_bins
        
    # Flatten the arrays for easier indexing
    mean_flat = mean_frame.flatten()
    frame_in_flat = frame_in.flatten()
        
    DN_val = 0
    for m in range(len(row_bins) - 1):
        for n in range(len(col_bins) - 1):
            DN_val += DN_bin
                
            # Create masks based on DN values
            mean_idx = (mean_flat >= DN_val) & (mean_flat < (DN_val + DN_bin))
                
            # Ensure that there is at least one element to calculate mean and std
            if np.any(mean_idx):
                local_mean_array[m, n] = np.nanmean(frame_in_flat[mean_idx])
                local_noise_array[m, n] = np.nanstd(frame_in_flat[mean_idx])
            else:
                local_mean_array[m, n] = 0
                local_noise_array[m, n] = 0
        
    return local_mean_array, local_noise_array

def diff2std(diff_frame, detector_regions=None):
    """
    calculate the standard deviation of the frame difference, 
    diff_frame within the ROI row and column boundaries.
    
    Args:
      diff_frame (np.array): frame diffference
      detector_regions (dict): a dictionary of detector geometry properties.
        Keys should be as found in detector_areas in detector.py.  Defaults to
        that dictionary. 
      
    Returns:
      np.array: standard deviation of frame difference
    """
        
    selected_area = slice_section(diff_frame, 'SCI', 'prescan_reliable',
        detector_regions)
    std_value = np.std(selected_area.reshape(-1), ddof=1)
    # dividing by sqrt(2) since we want std of one frame
    return std_value / np.sqrt(2)
    
def gauss(x, A, mean, sigma):
    """
    Gauss function. Called by sgaussfit.
    
    Args:
      x (np.array): input array
      A (float): amplitude
      mean (float): mean value
      sigma (float): sigma
    
    Returns:
      np.array: Gauss function
    """
    return A * np.exp(-0.5 * ((x - mean) ** 2) / (sigma ** 2))
    
def sgaussfit(xdata, ydata, gaussinp):
    """
    Find fitting parameters to Gauss function. Called by Single_peakfit. 
    gaussinp is an array containing initial values of A, mean, sigma.
    
    Args:
      xdata (np.array): input x array
      ydata (np.array): input y array
      gaussinp (np.array): initial guess array
    
    Returns:
      np.array: fit parameters
      np.array: fit parameters
    """
    popt, pcov = curve_fit(gauss, xdata, ydata, p0=gaussinp)
    sse = np.sum((ydata - gauss(xdata, *popt)) ** 2)
    return popt, lambda params: (sse, gauss(xdata, *params))
    
def Single_peakfit(xdata, ydata):
    """
    Fit Gauss function to x, y data. Returns mean and sigma. Only sigma 
    is used in main code call.
    
    Args:
      xdata (np.array): x data
      ydata (np.array): y data
    
    Returns:
      float: sigma
    """
    astart = np.max(ydata)
    mustart = xdata[np.argmax(ydata)]
    sigmastart = 10
    sgaussinp = [astart, mustart, sigmastart]
    
    estimates, model = sgaussfit(xdata, ydata, sgaussinp)
    a1, mu1, sigma = estimates

    return sigma
    
def histc_roi(frame, rn_bins, detector_regions=None):
    """
    Histogram of pixel values of frame within the ROI and bins defined in 
    rn_bins. Returns the counts in each bin.
    
    Args:
      frame (np.array): frame
      rn_bins (int): histogram bins
      detector_regions (dict): a dictionary of detector geometry properties.
        Keys should be as found in detector_areas in detector.py.  Defaults to
        that dictionary.      
    
    Returns:
      np.array: counts in each bin
    """
    selected_area = slice_section(frame, 'SCI', 'prescan_reliable',
        detector_regions)
    data_reshaped = selected_area.ravel()
    counts, _ = np.histogram(data_reshaped, bins=rn_bins)
      
    return counts
    
def calculate_mode(arr):
    """
    calculates histogram of an array
    
    Args:
      arr (np.array): input array
    
    Returns:
      np.array: bin center values
      np.array: bin center counts
    """
    counts, bin_edges = np.histogram(arr)
    # Calculate bin centers (values)
    values = (bin_edges[:-1] + bin_edges[1:]) / 2
    max_count_index = np.argmax(counts)
    return values[max_count_index], counts[max_count_index]

def sigma_clip(data, sigma=2.5, max_iters=6):
    """
    Perform sigma-clipping on the data.
      
    Args:
       data (np.array): The input data to be sigma-clipped.
       sigma (float): The number of standard deviations to use for clipping.
       max_iters (int): The maximum number of iterations to perform.
    
    Returns:
      np.ndarray: The sigma-clipped data.
      np.ndarray: A boolean mask where True indicates a clipped value.
    """
    data = np.asarray(data)
    clipped_data = data.copy()
      
    for i in range(max_iters):
        mean = np.mean(clipped_data)
        std = np.std(clipped_data)
        mask = np.abs(clipped_data - mean) > sigma * std
        if not np.any(mask):
            break
        clipped_data = clipped_data[~mask]
        
    return clipped_data, mask
    
######################### start of main code #############################

def calibrate_kgain(dataset_kgain, 
                    n_cal=10, n_mean=30, min_val=800, max_val=3000, binwidth=68,
                    make_plot=True,plot_outdir='figures', show_plot=False,
                    logspace_start=-1, logspace_stop=4, logspace_num=200,
                    verbose=False, detector_regions=None, kgain_params=None):
    """
    kgain (e-/DN) is calculated from the means and variances
    within the defined minimum and maximum mean values. A photon transfer curve
    is plotted from the std dev and mean values from the bins. 

    Args:
      dataset_kgain (corgidrp.Dataset): The frames in the dataset are
        bias-subtracted. The dataset contains frames belonging to two different
        sets -- Mean frame and a large array of unity gain frames.
        Mean frame: Unity gain frames with constant exposure time. These frames
        are used to create a mean pupil image. The mean frame is used to select
        pixels in each frame of the large array of unity gain frames (see next)
        to calculate its mean signal. In general, it is expected that at least
        30 frames or more will be taken for this set. In TVAC, 30 frames, each
        with an exposure time of 5.0 sec were taken.
        Large array of unity gain frames: Set of unity gain frames with subsets
        of equal exposure times. Data for each subset should be taken sequentially:
        Each subset must have at least 5 frames. All frames for a subset are taken
        before moving to the next subset. Two of the subsets have the same (repeated)
        exposure time. These two subsets are not contiguous: The first subset is
        taken near the start of the data collection and the second one is taken
        at the end of the data collection (see TVAC example below). The mean
        signal of these two subsets is used to correct for illumination
        brightness/sensor sensitivity drifts for all the frames in the whole set,
        depending on when the frames were taken. There should be no other repeated
        exposure time among the subsets. In TVAC, a total of 110 frames were taken
        within this category. The 110 frames consisted of 22 subsets, each with
        5 frames. All 5 frames had the same exposure time. The exposure times in
        TVAC in seconds were, each repeated 5 times to collect 5 frames in each
        subset -- 0.077, 0.770, 1.538, 2.308, 3.077, 3.846, 4.615, 5.385, 6.154,
        6.923, 7.692, 8.462, 9.231, 10.000, 11.538, 10.769, 12.308, 13.077,
        13.846, 14.615, 15.385, and 1.538 (again).
      n_cal (int):
        Minimum number of sub-stacks used to calibrate K-Gain. The default value
        is 10.
      n_mean (int):
        Minimum number of frames used to generate the mean frame. The default value
        is 30.
      min_val (int): 
        Minimum value (in DN) of mean values from sub-stacks to use in calculating 
        kgain. (> 400 recommended)  
      max_val (int):
        Maximum value (in DN) of mean values from sub-stacks to use in calculating 
        kgain. Choose value that avoids large deviations from linearity at high 
        counts. (< 6,000 recommended)
      binwidth (int):
        (Optional) Width of each bin for calculating std devs and means from each 
        sub-stack in dataset_kgain. Maximum value of binwidth is 800. Notice that small 
        values increase computation time.
        (minimum 10; binwidth between 65 and 75 recommended)
      make_plot (bool): (Optional) generate and store plots. Default is True.
      plot_outdir (str): (Optional) Output directory to store figues. Default is
        'figures'. The default directory is not tracked by git.
      show_plot (bool): (Optional) display the plots. Default is False.
      logspace_start (int): log plot min value in np.logspace.
      logspace_stop (int): log plot max value in np.logspace.
      logspace_num (int): Number of elements in np.logspace.
      verbose (bool): (Optional) display various diagnostic print messages.
        Default is False. 
      detector_regions (dict): a dictionary of detector geometry properties.
        Keys should be as found in detector_areas in detector.py.  Defaults to
        that dictionary.
      kgain_params (dict): (Optional) Dictionary containing row and col specifications
        for the region of interest (indicated by 'rowroi1','rowroi2','colroi1',and 'colroi2').
        The 'roi' needs one square region specified, and 'back' needs two square regions, 
        where a '1' ending indicates the smaller of two values, and a '2' ending indicates the larger 
        of two values.  The coordinates of the square region are specified by matching 
        up as follows: (rowroi1, colroi1), (rowroi2, colroi1), etc. 
        Also must contain:
        'rn_bins1': lower bound of counts histogram for fitting or read noise
        'rn_bins2': upper bound of counts histogram for fitting or read noise 
        'max_DN_val': maximum DN value to be included in photon transfer curve (PTC)
        'signal_bins_N': number of bins in the signal variables of PTC curve
        Defaults to kgain_params_default included in this file.
    
    Returns:
      corgidrp.data.KGain: kgain estimate from the least-squares fit to the photon
        transfer curve (in e-/DN). The expected value of kgain for EXCAM with
        flight readout sequence should be between 8 and 9 e-/DN
    """
    if kgain_params is None:
        kgain_params = kgain_params_default
        
    check_kgain_params(kgain_params)

    if detector_regions is None:
        detector_regions = detector_areas

    # cast dataset objects into np arrays for convenience
    cal_list, mean_frame_list, actual_gain = kgain_dataset_2_list(dataset_kgain)

    # check number of frames, unique EM value, exposure times and datetimes
    tmp = cal_list[0]
    for idx in range(4):
        try:
            tmp = tmp[idx]
        except:
            pass
    if idx != 3:
        raise CalKgainException('cal_list must be 4-D (i.e., a stack of '
                '3-D sub-stacks)')
    if len(cal_list) < n_cal:
        raise CalKgainException(f'Number of sub-stacks in cal_list must '
                'be more than {n_cal}.')
    if len(cal_list) < 20 :
        warnings.warn('Number of sub-stacks in cal_list is less than 20, '
        'which is the recommended minimum number for a good fit ')
    for i in range(len(cal_list)):
        check.threeD_array(cal_list[i], 'cal_list['+str(i)+']', TypeError)
        if len(cal_list[i]) < 5:
            raise CalKgainException('A sub-stack in cal_list was found with less than 5 '
            'frames, which is the required minimum number per sub-stack')
        if i > 0:
            if len(cal_list[i-1]) != len(cal_list[i]):
                raise CalKgainException('All sub-stacks must have the '
                            'same number of frames and frame shape.')
    tmp = mean_frame_list[0]
    for idx in range(3):
        try:
            tmp = tmp[idx]
        except:
            pass
    if idx != 2:    
        raise CalKgainException('mean_frame_list must be 3-D (i.e., a stack of '
                '2-D sub-stacks')
    if len(mean_frame_list) < n_mean:
        raise CalKgainException(f'Number of sub-stacks in mean_frame_list must '
                'be equal to or greater than {n_mean}.')

    check.real_positive_scalar(actual_gain, 'actual_gain', TypeError)
    if actual_gain < 1:
        raise CalKgainException('Actual gain must be >= 1.')
    check.positive_scalar_integer(min_val, 'min_val', TypeError)
    check.positive_scalar_integer(max_val, 'max_val', TypeError)
    if min_val >= max_val:
        raise CalKgainException('max_val must be greater than min_val')
    check.positive_scalar_integer(binwidth, 'binwidth', TypeError)
    if binwidth < 10:
        raise CalKgainException('binwidth must be >= 10.')
    if binwidth > 800:
        raise CalKgainException('binwidth must be < 800.')
    if not isinstance(logspace_start, (float, int)):
        raise TypeError('logplot1 is not a number')
    if not isinstance(logspace_stop, (float, int)):
        raise TypeError('logplot2 is not a number')
    if not isinstance(logspace_num, (float, int)):
        raise TypeError('logplot3 is not a number')
    
    # get relevant constants
    rowroi1 = kgain_params['rowroi1']
    rowroi2 = kgain_params['rowroi2']
    colroi1 = kgain_params['colroi1']
    colroi2 = kgain_params['colroi2']
    rn_bins1 = kgain_params['rn_bins1']
    rn_bins2 = kgain_params['rn_bins2']
    max_DN_val = kgain_params['max_DN_val']
    signal_bins_N = kgain_params['signal_bins_N']

    if make_plot is True:
        # Avoid issues with importing matplotlib on headless servers without GUI
        # support without proper configuration
        import matplotlib.pyplot as plt
        # Output directory
        if os.path.exists(plot_outdir) is False:
            os.mkdir(plot_outdir)
            if verbose:
                print('Output directory for figures created in ', os.getcwd())
    
    # NOTE: binwidth must be <= rowroi and colroi
    rowroi = slice(rowroi1,rowroi2)
    colroi = slice(colroi1,colroi2)
    
    averages = []
    deviations = []
    read_noise = []
    deviations_shot = []
    #mean_signals = []
    
    rn_bins = np.array(range(rn_bins1, rn_bins2))
    bin_locs = rn_bins[0:-1]

    max_DN = max_DN_val; # maximum DN value to be included in PTC
    nrow = len(cal_list[0][0])
    ncol = len(cal_list[0][0][0])
    
    # prepare "good mean frame"
    good_mean_frame = np.zeros((nrow, ncol))
    nFrames2 = len(mean_frame_list)
    for mean_frame_count in range(nFrames2):
        good_mean_frame += mean_frame_list[mean_frame_count]
    
    good_mean_frame = good_mean_frame / nFrames2
    
    # Slice the good_mean_frame array
    frame_slice = good_mean_frame[rowroi, colroi]
    
    # If requested, create a figure and plot the sliced frame
    if make_plot:
        fname = 'kgain_mean_frame'
        plt.figure()
        # 'viridis' is a common colormap
        plt.imshow(frame_slice, aspect='equal', cmap='viridis')
        plt.colorbar()
        plt.title('Good quality mean frame')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    nFrames = len(cal_list[0]) # number of frames in an exposure set
    nSets = len(cal_list) # number of exposure sets
    
    # Start with specific offset indices for list comprehension in jj loop
    index_offsets = [2, 3, 1, 4, 5]
    
    # Start with specific index pairs for list comprehension in jj loop
    index_pairs = [(0, 1), (0, 2), (1, 2), (2, 3), (3, 4)]
    # if nFrames>5, then append additional terms to index_pairs
    if nFrames > 5:
        # append offsets
        index_offsets += [i for i in range(4, nFrames - 1)]
        
        # append index pairs (i, i + 1) for i from 4 to len(frames) - 2
        index_pairs += [(i, i + 1) for i in range(4, nFrames - 1)]
    
    for jj in range(nSets):
        if verbose:
            print(jj)
        
        # multi-frame analysis method
        frames = [cal_list[jj][nFrames - offset] for offset in index_offsets]
        # Calculate frame differences
        frames_diff = [frames[j] - frames[k] for j, k in index_pairs]
        # calculate read noise with std from prescan
        rn_std = [diff2std(frames_diff[x], detector_regions) for x 
            in range(len(frames_diff))]
        
        counts_diff = [histc_roi(frames_diff[x], rn_bins, detector_regions) for x 
            in range(len(frames_diff))]
        
        # calculate read noise from prescan with Gaussian fit
        rn_gauss = [Single_peakfit(bin_locs,counts_diff[x])/np.sqrt(2) for x 
            in range(len(frames_diff))]
        
        # split each frame up into bins, take std and mean of each region
        mean_frames_mean_curr0 = np.mean(frames, axis=0)
        mean_frames_mean_curr = np.mean(mean_frames_mean_curr0[rowroi,colroi])
        
        # Calculate the means
        mean_good_mean_frame_roi = np.mean(good_mean_frame[rowroi,colroi])
    
        # Compute the scaling factor
        scaling_factor = mean_frames_mean_curr / mean_good_mean_frame_roi
    
        # Apply the scaling to the selected ROI in good_mean_frame
        good_mean_frame_scaled = scaling_factor * good_mean_frame[rowroi,colroi]
        
        # calculate means and stdevs using good_mean_frame_scaled logical indexing
        local_mean_arrays = [ptc_bin2(frame[rowroi, colroi], 
                            good_mean_frame_scaled, 
                            binwidth, max_DN)[0] for frame in frames]
        
        local_noise_arrays = [ptc_bin2(frame[rowroi, colroi], 
                            good_mean_frame_scaled, 
                            binwidth, max_DN)[1] for frame in frames_diff]
        std_diffs = local_noise_arrays / np.sqrt(2)
        
        averages0 = [array.reshape(-1, 1) for array in local_mean_arrays]
        averages.extend(averages0)
        deviations0 = [array.reshape(-1, 1) for array in std_diffs]
        deviations.extend(deviations0)
        
        added_deviations_shot_arr = [
                np.sqrt(np.square(np.reshape(std_diffs[x], 
                newshape=(-1, 1))) - complex(rn_std[x])**2)
                for x in range(len(rn_std))
                ]
        
        deviations_shot.extend(added_deviations_shot_arr)

        read_noise.extend(rn_std)
        
    averages = np.concatenate([arr.flatten() for arr in averages])
    deviations = np.concatenate([arr.flatten() for arr in deviations])
    deviations_shot = np.concatenate([arr.flatten() for arr in deviations_shot])
    averages_deviations_vector = np.column_stack((averages, np.abs(deviations_shot)))
    
    # Generate linearly spaced bins
    signal_bins = np.linspace(np.min(averages), np.max(averages), signal_bins_N)
    signal_bins = np.insert(signal_bins, 0, 0)  # Insert 0 at the beginning
    
    # Initialize containers for the results
    binned_averages_compiled = []
    binned_averages_error = []
    binned_shot_deviations_compiled = []
    binned_deviations_error = []
    binned_total_deviations = []

    # Loop through each bin, excluding the first which is just 0
    for b in range(1, len(signal_bins)):
        average_bool = (averages_deviations_vector[:, 0] > signal_bins[b-1]) & (averages_deviations_vector[:, 0] < signal_bins[b])
    
        # Filter data within the current bin
        current_binned_averages = averages_deviations_vector[:, 0][average_bool]
        current_binned_deviations = averages_deviations_vector[:, 1][average_bool]
        current_binned_total_deviations = deviations[average_bool]
    
        # Compute statistics if there are data points within the bin
        if current_binned_averages.size > 0:
            binned_averages_compiled.append(np.mean(current_binned_averages))
            binned_averages_error.append(np.std(current_binned_averages, ddof=1) / np.sqrt(current_binned_averages.size))
            binned_shot_deviations_compiled.append(np.mean(current_binned_deviations))
            binned_deviations_error.append(np.std(current_binned_deviations, ddof=1) / np.sqrt(current_binned_averages.size))
            binned_total_deviations.append(np.mean(current_binned_total_deviations))
        else:
            # Append NaN or some other placeholder if no data points in the bin
            binned_averages_compiled.append(np.nan)
            binned_averages_error.append(np.nan)
            binned_shot_deviations_compiled.append(np.nan)
            binned_deviations_error.append(np.nan)
            binned_total_deviations.append(np.nan)

    # Convert lists to numpy arrays
    compiled_binned_averages = np.array(binned_averages_compiled)
    compiled_binned_shot_deviations = np.array(binned_shot_deviations_compiled)
    compiled_binned_total_deviations = np.array(binned_total_deviations)
    
    compiled_binned_deviations_error = np.array(binned_deviations_error)
    
    # define bounds for linearity fit
    # lower bound indices should include as many data points as possible
    # upper bound indices should avoid nonlin region
    lower_bound = np.min(np.where(compiled_binned_averages > min_val)[0])
    upper_bound = np.max(np.where(compiled_binned_averages < max_val)[0])
    
    # Logarithmic transformation of specific array segments
    logged_averages = np.log10(compiled_binned_averages[lower_bound:upper_bound])
    logged_deviations = np.log10(compiled_binned_shot_deviations[lower_bound:upper_bound])
    
    # Find indices of NaN's
    nan_dev = np.isnan(logged_deviations)
    nan_avg = np.isnan(logged_averages)
    
    # Filter out NaN's from the data
    logged_deviations = logged_deviations[~nan_dev & ~nan_avg]
    logged_averages = logged_averages[~nan_dev & ~nan_avg]
    
    # Calculate the error factor
    logged_deviations_error_factor = compiled_binned_deviations_error / compiled_binned_shot_deviations
    logged_deviations_error_factor = logged_deviations_error_factor[lower_bound:upper_bound]
    logged_deviations_error_factor = logged_deviations_error_factor[~nan_dev & ~nan_avg]
    
    # Calculate the logged deviations error
    logged_deviations_error = logged_deviations_error_factor * logged_deviations
    
    # Create a mask for non-NaN values
    non_nan_mask = ~np.isnan(logged_deviations_error)

    # Use the mask to filter out NaN values
    y_err = logged_deviations_error[non_nan_mask]
    x_vals = logged_averages[non_nan_mask]
    y_vals = logged_deviations[non_nan_mask]
    
    # make array of kgain values from variance and mean values
    # equation: k_gain = mean / signal variance
    kgain_arr = 10**(x_vals)/(10**y_vals)**2
    
    if make_plot:
        fname = 'kgain_histogram'
        plt.figure()
        # 'auto' lets matplotlib decide the number of bins
        plt.hist(kgain_arr, bins='auto', log=True)
        plt.title('Histogram of kgain values')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    if make_plot:
        fname = 'kgain_vs_mean'
        plt.figure()
        plt.plot(10**x_vals, kgain_arr, marker='o', linestyle='-', color='b', label='kgain')
        # Set y-axis range
        plt.ylim(6, 11)
        # Add a horizontal line at y = 10
        plt.axhline(y=8.7, color='r', linestyle='--', label='y = 8.7')
        # Add a grid
        plt.grid(True)
        plt.title('kgain versus mean')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    kgain_clipped, _ = sigma_clip(kgain_arr)
    mode_kgain,_ = calculate_mode(kgain_clipped)

    # adopt 'mode_kgain' as the final value to return
    kgain = mode_kgain

    if make_plot:
        fname = 'kgain_clipped_histogram'
        plt.figure()
        # 'auto' lets matplotlib decide the number of bins
        plt.hist(kgain_clipped, bins='auto', log=True)
        plt.title('Histogram of clipped kgain values')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    # parameter for plot
    parm1 = -0.5*np.log10(kgain)
    
    # Gaussian read noise value in DN
    mean_rn_gauss_DN = np.mean(rn_gauss)
    mean_rn_gauss_e = mean_rn_gauss_DN * kgain
    
    mean_rn_std_DN = np.mean(read_noise)
    mean_rn_std_e = mean_rn_std_DN * kgain
    
    # If requested, plotting
    if make_plot:
        fname = 'kgain_ptc'
        # Create log-spaced averages for plotting
        full_range_averages = np.logspace(logspace_start, logspace_stop, logspace_num)
        
        plt.figure(num=3, figsize=(20, 10))
        
        # Main data plots
        plt.plot(compiled_binned_averages, compiled_binned_total_deviations, 
                 label='Shot Noise + Read Noise')
        # Assuming `lower_bound` is 6
        plt.plot(compiled_binned_averages, 
                 compiled_binned_shot_deviations, label='Shot Noise')
        plt.errorbar(10**logged_averages, 10**logged_deviations, 
                     yerr=10**(logged_deviations_error), color='gray', 
                     label='Fitted Points')
        
        # Extra lines
        plt.plot(10**full_range_averages, 
                 10**(full_range_averages * 0.50 + parm1), 
                 'k-', label='Forced 0.50 Gradient')
        
        # Reference lines and text annotations
        plt.axhline(y=mean_rn_std_DN, color='k', linestyle='--')
        plt.axvline(x=95000, color='k', linestyle='--')
        plt.text(2, 1.5, f'Fitted k gain = {np.round(kgain,1)} e/DN', fontsize=15)
        plt.text(2, 10, 
                 f'Read noise (Gaussian fit) = {np.round(mean_rn_gauss_DN,1)} DN = {mean_rn_gauss_e} e', 
                 fontsize=15)
        plt.text(2, 8, f'Read noise (Std) = {np.round(mean_rn_std_DN,1)} DN = {np.round(mean_rn_std_e,1)} e', 
                 fontsize=15)
        
        # Scaling and labeling
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('Signal (DN)')
        plt.ylabel('Noise (DN)')
        plt.title('PTC curve fit')
        plt.grid(True)
        
        # Setting axis limits
        plt.xlim([1, 20000])
        plt.ylim([1, 1000])
        
        # Legend and aesthetics
        plt.legend(loc='upper left')
        
        plt.gca().tick_params(axis='both', which='major', labelsize=18)
        
        # Adjust figure position
        plt.gcf().set_dpi(100)
        plt.gcf().set_size_inches(20, 10)
        
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    # prepare PTC array
    ptc_list = [compiled_binned_averages, 
             compiled_binned_shot_deviations]
    ptc = np.column_stack(ptc_list)

    prhd = dataset_kgain.frames[0].pri_hdr
    exthd = dataset_kgain.frames[0].ext_hdr
    # Read noise and error
    exthd['RN'] = mean_rn_gauss_e
    
    # rn err depends on spread of data that determines rn and the error in kgain,
    # so use error propagation to find error in (rn in DN)*kgain
    kgain_err = np.std(kgain_clipped)
    rn_err_DN = np.std(rn_gauss)
    rn_err_e = np.sqrt((kgain*rn_err_DN)**2 + (mean_rn_gauss_DN*kgain_err)**2)
    exthd['RN_ERR'] = rn_err_e
    exthd['RN_UNIT'] = 'detected electrons'
    
    # Update history
    exthd['HISTORY'] = f"Kgain and read noise derived from a set of frames on {exthd['DATETIME']}"
    gain_value = np.array([[kgain]])

    kgain = data.KGain(gain_value, err = np.array([[np.std(kgain_clipped)]]), ptc = ptc, pri_hdr = prhd, ext_hdr = exthd, input_dataset=dataset_kgain)
    
    return kgain

def kgain_dataset_2_list(dataset):
    """
    Casts the CORGIDRP Dataset object for K-gain calibration into a list of
    numpy arrays sharing the same exposure time. It also returns the list of
    unique EM values and set of exposure times used with each EM. Note: EM gain
    is the commanded values: EMGAIN_C.

    This function also performs a set of tests about the data type and values in
    dataset.

    Args:
        dataset (corgidrp.Dataset): Dataset with a set of of EXCAM illuminated
        pupil L1 SCI frames (counts in DN)

    Returns:
        list with stack of stacks of data array associated with each frame
        array of exposure times associated with each frame
        array of datetimes associated with each frame

    """
    # Split Dataset
    dataset_cp = dataset.copy()
    split = dataset_cp.split_dataset(exthdr_keywords=['EXPTIME'])

    # Data
    stack = []
    # Mean frame data
    mean_frame_stack = []
    # Exposure times
    exp_times = []
    # Datetimes
    datetimes = []
    # EM gains (There can only be an EM gain in the data used to calibrate K-gain)
    em_gains = []
    # Size of each sub stack
    len_sstack = []
    # Record measured gain of each substack of calibration frames
    gains = []    
    for idx_set, data_set in enumerate(split[0]):
        # Second layer (array of different exposure times)
        sub_stack = []
        record_exp_time = True
        record_len = True
        record_gain = True
        for frame in data_set.frames:
            if record_exp_time:
                exp_time_mean_frame = frame.ext_hdr['EXPTIME']
                record_exp_time = False

            if frame.ext_hdr['EXPTIME'] != exp_time_mean_frame:
                raise Exception('Frames in the same data set must have the same exposure time')

            if frame.pri_hdr['OBSNAME'] == 'MNFRAME':
                if frame.ext_hdr['EMGAIN_C'] != 1:
                    raise Exception('The commanded gain used to build the mean frame must be unity')
                mean_frame_stack.append(frame.data)
            else:
                if record_len:
                    len_sstack.append(len(data_set.frames))
                    record_len = False
                sub_stack.append(frame.data)
                exp_time = frame.ext_hdr['EXPTIME']
                if isinstance(exp_time, float) is False:
                    raise Exception('Exposure times must be float')
                if exp_time <=0:
                    raise Exception('Exposure times must be positive')
                exp_times.append(exp_time)
                datetime = frame.ext_hdr['DATETIME']
                if isinstance(datetime, str) is False:
                    raise Exception('DATETIME must be a string')
                datetimes.append(datetime)
                em_gain = frame.ext_hdr['EMGAIN_C']
                if em_gain < 1:
                    raise Exception('Commanded EM gain must be >= 1')
                em_gains.append(em_gain)
                if record_gain:
                    try: # if EM gain measured directly from frame TODO change hdr name if necessary
                        gains.append(frame.ext_hdr['EMGAIN_M'])
                    except:
                        if frame.ext_hdr['EMGAIN_A'] > 0: # use applied EM gain if available
                            gains.append(frame.ext_hdr['EMGAIN_A'])
                        else: # use commanded gain otherwise
                            gains.append(frame.ext_hdr['EMGAIN_C'])
                        record_gain = False
                
        # Calibration data may have different subsets
        if len(sub_stack) != 0:
            stack.append(np.stack(sub_stack))

    # Need to split substacks with the same exposure times
    stack_cp = []
    # Get expected size of substacks
    len_sub = min(len_sstack)
    # Length of substack must be at least 1
    if len(len_sstack) == 0:
        raise Exception('Substacks must have at least one element')
    for sub in stack:
        if len(sub) == len_sub:
            stack_cp.append(sub)
        else:
            # Add extra care confirming all collected substacks have same # frames
            if len(sub)/len_sub != len(sub)//len_sub:
                raise Exception('All substacks must have the same number of frames')
            idx_0 = 0
            for rep in range(len(sub)//len_sub):
                stack_cp.append(sub[idx_0:idx_0+len_sub])
                idx_0 += len_sub
    stack = stack_cp        
    # All elements of datetimes must be unique
    if len(datetimes) != len(set(datetimes)):
        raise Exception('DATETIMEs cannot be duplicated')
    # There can only be an EM gain in the data used to calibrate K-gain
    if len(set(em_gains)) != 1:
        raise Exception('There can only be one commanded EM gain when calibrating K-Gain')
    if np.any(np.array(gains) < 1):
        raise Exception('Actual EM gains must be greater than or equal to 1')
    # When measuring k_gain, there can only be one gain for all exposure times
    actual_gain = np.mean(gains) # not actually used in k gain calibration since frames already gain-divided
    
    return stack, mean_frame_stack, actual_gain



# Place to put detector-related utility functions

import numpy as np
from scipy import interpolate
from scipy.ndimage import median_filter
from scipy.ndimage import gaussian_filter as gauss
from scipy import ndimage
from scipy.signal import convolve2d
import astropy.io.fits as fits
from astropy.convolution import convolve_fft
import photutils.centroids as centr
from photutils.aperture import CircularAperture

import corgidrp.data as data

def get_relgains(frame, em_gain, non_lin_correction):
    """
    For a given bias subtracted frame of dn counts, return a same sized
    array of relative gain values.

    This algorithm contains two interpolations:

    - A 2d interpolation to find the relative gain curve for a given EM gain
    - A 1d interpolation to find a relative gain value for each given dn
      count value.

    Both of these interpolations are linear, and both use their edge values as
    constant extrapolations for out of bounds values.

    Parameters:
        frame (array_like): Array of dn count values.
        em_gain (float): Detector EM gain.
        non_lin_correction (corgi.drp.NonLinearityCorrection): A NonLinearityCorrection calibration file.

    Returns:
        array_like: Array of relative gain values.
    """
    if non_lin_correction is None: # then no correction
        return np.ones_like(frame) 
    
    # Column headers are gains, row headers are dn counts
    gain_ax = non_lin_correction.data[0, 1:]
    count_ax = non_lin_correction.data[1:, 0]
    # Array is relative gain values at a given dn count and gain
    relgains = non_lin_correction.data[1:, 1:]

    #MMB Note: This check is maybe better placed in the code that is saving the non-linearity correction file?
    # Check for increasing axes
    if np.any(np.diff(gain_ax) <= 0):
        raise ValueError('Gain axis (column headers) must be increasing')
    if np.any(np.diff(count_ax) <= 0):
        raise ValueError('Counts axis (row headers) must be increasing')
    # Check that curves (data in columns) contain or straddle 1.0
    if (np.min(relgains, axis=0) > 1).any() or \
       (np.max(relgains, axis=0) < 1).any():
        raise ValueError('Gain curves (array columns) must contain or '
                              'straddle a relative gain of 1.0')

    # Create interpolation for em gain (x), counts (y), and relative gain (z).
    # Note that this defaults to using the edge values as fill_value for
    # out of bounds values (same as specified below in interp1d)
    f = interpolate.RectBivariateSpline(gain_ax,
                                    count_ax,
                                    relgains.T,
                                    kx=1,
                                    ky=1,
    )
    # Get the relative gain curve for the given gain value
    relgain_curve = f(em_gain, count_ax)[0]

    # Create interpolation for dn counts (x) and relative gains (y). For
    # out of bounds values use edge values
    ff = interpolate.interp1d(count_ax, relgain_curve, kind='linear',
                              bounds_error=False,
                              fill_value=(relgain_curve[0], relgain_curve[-1]))
    # For each dn count, find the relative gain
    counts_flat = ff(frame.ravel())

    return counts_flat.reshape(frame.shape)

detector_areas= {
    'SCI' : {
        'frame_rows' : 1200,
        'frame_cols' : 2200,
        'image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [13, 1088]
            },
        'prescan' : {
            'rows': 1200,
            'cols': 1088,
            'r0c0': [0, 0],
            'col_start': 800,
            'col_end': 1000,
            },
        'prescan_reliable' : {
            'rows': 1200,
            'cols': 200,
            'r0c0': [0, 800]
            },
        'parallel_overscan' : {
            'rows': 163,
            'cols': 1056,
            'r0c0': [1037, 1088]
            },
        'serial_overscan' : {
            'rows': 1200,
            'cols': 56,
            'r0c0': [0, 2144]
            },
        },
    'ENG' :{
        'frame_rows' : 2200,
        'frame_cols' : 2200,
        'image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [13, 1088]
            },
        'prescan' : {
            'rows': 2200,
            'cols': 1088,
            'r0c0': [0, 0],
            'col_start': 800,
            'col_end': 1000,
            },
        'prescan_reliable' : {
            'rows': 2200,
            'cols': 200,
            'r0c0': [0, 800]
            },
        'parallel_overscan' : {
            'rows': 1163,
            'cols': 1056,
            'r0c0': [1037, 1088]
            },
        'serial_overscan' : {
            'rows': 2200,
            'cols': 56,
            'r0c0': [0, 2144]
            },
        },
    'ENG_EM' :{
        'frame_rows' : 2200,
        'frame_cols' : 2200,
        'image' : { # combined lower and upper
            'rows': 2048,
            'cols': 1024,
            'r0c0': [13, 1088]
            },
        'lower_image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [13, 1088]
            },
        'upper_image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [1037, 1088]
            },
        'prescan' : {
            'rows': 2200,
            'cols': 1088,
            'r0c0': [0, 0],
            'col_start': 800,
            'col_end': 1000
            },
        'prescan_reliable' : {
            'rows': 2200,
            'cols': 200,
            'r0c0': [0, 800]
            },
        'parallel_overscan' : {
            'rows': 130,
            'cols': 1056,
            'r0c0': [2070, 1088]
            },
        'serial_overscan' : {
            'rows': 2200,
            'cols': 56,
            'r0c0': [0, 2144]
            },
        },
    'ENG_CONV' :{
        'frame_rows' : 2200,
        'frame_cols' : 2200,
        'image' : { # combined lower and upper
            'rows': 2048,
            'cols': 1024,
            'r0c0': [13, 48]
            },
        'lower_image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [13, 48]
            },
        'upper_image' : {
            'rows': 1024,
            'cols': 1024,
            'r0c0': [1037, 48]
            },
        # 'prescan' is actually the serial_overscan region, but the code needs to take
        # the bias from the largest serial non-image region, and the code identifies
        # this region as the "prescan", so we have the prescan and serial_overscan
        # names flipped for this reason.
        'prescan' : {
            'rows': 2200,
            'cols': 1128,
            'r0c0': [0, 1072],
            'col_start': 1200,
            'col_end': 1400
            },
        'prescan_reliable' : {
            # not sure if these are good in the eng_conv case where the geometry is
            # flipped relative to the other cases, but these cols would where the
            # good, reliable cols used for getting row-by-row bias
            # would be
            'rows': 2200,
            'cols': 200,
            'r0c0': [0, 1200]
            },
        'parallel_overscan' : {
            'rows': 130,
            'cols': 1056,
            'r0c0': [2070, 16]
            },
        'serial_overscan' : {
            'rows': 2200,
            'cols': 16,
            'r0c0': [0, 0]
            },
        }
    }


def slice_section(frame, arrtype, key, detector_regions=None):

    """
    Slice 2d section out of frame

    Ported from II&T read_metadata.py

    Args:
        frame (np.ndarray): Full frame consistent with size given in frame_rows, frame_cols
        arrtype (str): Keyword referencing the observation type (e.g. 'ENG' or 'SCI')
        key (str): Keyword referencing section to be sliced; must exist in detector_areas
        detector_regions (dict): a dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py.  Defaults to that dictionary.

    Returns:
        np.ndarray: a 2D array of the specified detector area
    """
    if detector_regions is None:
            detector_regions = detector_areas
    rows = detector_regions[arrtype][key]['rows']
    cols = detector_regions[arrtype][key]['cols']
    r0c0 = detector_regions[arrtype][key]['r0c0']

    section = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
    if section.size == 0:
        raise Exception('Corners invalid. Tried to slice shape of {0} from {1} to {2} rows and {3} columns'.format(frame.shape, r0c0, rows, cols))
    return section

def embed(data, arrtype, key, pad_val=0, detector_regions=None):
        '''
        Embed subframe data into a full frame with a specified value per pixel everywhere else. 
        
        Args:
            data (np.ndarray):  Subframe data to embed
            arrtype (str): Keyword referencing the observation type (e.g. 'ENG' or 'SCI')
            key (str): Keyword referencing section to be sliced; must exist in detector_areas
            pad_val (float): Value to fill in each pixel outside the subframe region
            detector_regions (dict): a dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py.  Defaults to that dictionary.

        Returns:
            np.ndarray: a 2D array of the full detector area with embedded subframe
        '''
        if detector_regions is None:
            detector_regions = detector_areas
        ff_rows = detector_regions[arrtype]['frame_rows']
        ff_cols = detector_regions[arrtype]['frame_cols']
        full_frame = pad_val*np.ones((ff_rows, ff_cols))
        rows, cols, r0c0 = unpack_geom(arrtype, key, detector_regions)
        try:
            full_frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols] = data
        except Exception:
            raise Exception('Data does not fit in selected section')
        return full_frame

def unpack_geom(arrtype, key, detector_regions=None):
    """Safely check format of geom sub-dictionary and return values.

    Args:
        arrtype: str
        Keyword referencing the observation type (e.g. 'ENG' or 'SCI')
        key: str
        Desired section
        detector_regions: dict
        a dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py.  Defaults to that dictionary.

    Returns:
        rows: int
        Number of rows of frame
        cols : int
        Number of columns of frame
        r0c0: tuple
        Tuple of (row position, column position) of corner closest to (0,0)
    """
    if detector_regions is None:
        detector_regions = detector_areas
    coords = detector_regions[arrtype][key]
    rows = coords['rows']
    cols = coords['cols']
    r0c0 = coords['r0c0']

    return rows, cols, r0c0

def imaging_area_geom(arrtype, detector_regions=None):
    """Return geometry of imaging area (including shielded pixels)
    in reference to full frame.  Different from normal image area.

    Args:
        arrtype: str
        Keyword referencing the observation type (e.g. 'ENG' or 'SCI')
        detector_regions: dict
        a dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py.  Defaults to that dictionary.


    Returns:
        rows: int
        Number of rows of imaging area
        cols : int
        Number of columns of imaging area
        r0c0: tuple
        Tuple of (row position, column position) of corner closest to (0,0)
    """
    if detector_regions is None:
        detector_regions = detector_areas
    _, cols_pre, _ = unpack_geom(arrtype, 'prescan', detector_regions)
    _, cols_serial_ovr, _ = unpack_geom(arrtype, 'serial_overscan', detector_regions)
    rows_parallel_ovr, _, _ = unpack_geom(arrtype, 'parallel_overscan', detector_regions)
    #_, _, r0c0_image = self._unpack_geom('image')
    fluxmap_rows, _, r0c0_image = unpack_geom(arrtype, 'image', detector_regions)

    rows_im = detector_regions[arrtype]['frame_rows'] - rows_parallel_ovr
    cols_im = detector_regions[arrtype]['frame_cols'] - cols_pre - cols_serial_ovr
    r0c0_im = r0c0_image.copy()
    r0c0_im[0] = r0c0_im[0] - (rows_im - fluxmap_rows)

    return rows_im, cols_im, r0c0_im

def imaging_slice(arrtype, frame, detector_regions=None):
    """Select only the real counts from full frame and exclude virtual.
    Includes shielded pixels.

    Use this to transform mask and embed from acting on the full frame to
    acting on only the image frame.

    Args:
        arrtype: str
        Keyword referencing the observation type (e.g. 'ENG' or 'SCI')
        frame: array_like
        Input frame
        detector_regions: dict
        a dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py.  Defaults to that dictionary.

    Returns:
        sl: array_like
        Imaging slice

    """
    rows, cols, r0c0 = imaging_area_geom(arrtype, detector_regions)
    sl = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
    return sl

def flag_cosmics(cube, fwc, sat_thresh, plat_thresh, cosm_filter, cosm_box,
                   cosm_tail, mode='image', detector_regions=None, arrtype='SCI'):
    """Identify and remove saturated cosmic ray hits and tails.

    Use sat_thresh (interval 0 to 1) to set the threshold above which cosmics
    will be detected. For example, sat_thresh=0.99 will detect cosmics above
    0.99*fwc.

    Use plat_thresh (interval 0 to 1) to set the threshold under which cosmic
    plateaus will end. For example, if plat_thresh=0.85, once a cosmic is
    detected the beginning and end of its plateau will be determined where the
    pixel values drop below 0.85*fwc.

    Use cosm_filter to determine the smallest plateaus (in pixels) that will
    be identified. A reasonable value is 2.

    Args:
        cube (array_like, float):
            3D cube of image data (bias of zero).
        fwc (float):
            Full well capacity of detector *in DNs*.  Note that this may require a
            conversion as FWCs are usually specified in electrons, but the image
            is in DNs at this point.
        sat_thresh (float):
            Multiplication factor for fwc that determines saturated cosmic pixels.
        plat_thresh (float):
            Multiplication factor for fwc that determines edges of cosmic plateu.
        cosm_filter (int):
            Minimum length in pixels of cosmic plateus to be identified.
        cosm_box (int):
            Number of pixels out from an identified cosmic head (i.e., beginning of
            the plateau) to mask out.
            For example, if cosm_box is 3, a 7x7 box is masked,
            with the cosmic head as the center pixel of the box.
        cosm_tail (int):
            Number of pixels in the row downstream of the end of a cosmic plateau
            to mask.  If cosm_tail is greater than the number of
            columns left to the end of the row from the cosmic
            plateau, the cosmic masking ends at the end of the row.
        mode (string):
            If 'image', an image-area input is assumed, and if the input
            tail length is longer than the length to the end of the image-area row,
            the mask is truncated at the end of the row.
            If 'full', a full-frame input is assumed, and if the input tail length
            is longer than the length to the end of the full-frame row, the masking
            continues onto the next row.  Defaults to 'image'.
        detector_regions: (dict):  
            A dictionary of detector geometry properties.  Keys should be as 
            found in detector_areas in detector.py. Defaults to detector_areas in detector.py.
        arrtype (string):
            'ARRTYPE' from the header associated with the input data cube.  Defaults to 'SCI'.

    Returns:
        array_like, int:
            Mask for pixels that have been set to zero.

    Notes
    -----
    This algorithm uses a row by row method for cosmic removal. It first finds
    streak rows, which are rows that potentially contain cosmics. It then
    filters each of these rows in order to differentiate cosmic hits (plateaus)
    from any outlier saturated pixels. For each cosmic hit it finds the leading
    ledge of the plateau and kills the plateau (specified by cosm_filter) and
    the tail (specified by cosm_tail).

    |<-------- streak row is the whole row ----------------------->|
     ......|<-plateau->|<------------------tail---------->|.........

    B Nemati and S Miller - UAH - 02-Oct-2018
    Kevin Ludwick - UAH - 2024

    """
    mask = np.zeros(cube.shape, dtype=int)

    if detector_regions is None:
        detector_regions = detector_areas

    if mode=='full':
        im_num_rows = detector_regions[arrtype]['image']['rows']
        im_num_cols = detector_regions[arrtype]['image']['cols']
        im_starting_row = detector_regions[arrtype]['image']['r0c0'][0]
        im_ending_row = im_starting_row + im_num_rows
        im_starting_col = detector_regions[arrtype]['image']['r0c0'][1]
        im_ending_col = im_starting_col + im_num_cols
    else:
        im_starting_row = 0
        im_ending_row = mask.shape[1] - 1 # - 1 to get the index, not size
        im_starting_col = 0
        im_ending_col = mask.shape[2] - 1 # - 1 to get the index, not size

    # Do a cheap prefilter for rows that don't have anything bright
    max_rows = np.max(cube, axis=-1,keepdims=True)
    ji_streak_rows = np.transpose(np.array((max_rows >= sat_thresh*fwc).nonzero()[:-1]))

    for j,i in ji_streak_rows:
        row = cube[j,i]
        if i < im_starting_row or i > im_ending_row:
            continue
        # Find if and where saturated plateaus start in streak row
        i_begs = find_plateaus(row, fwc, sat_thresh, plat_thresh, cosm_filter)

        # If plateaus exist, kill the hit and the tail
        cutoffs = np.array([])
        ex_l = np.array([])
        if i_begs is not None:
            for i_beg in i_begs:
                if i_beg < im_starting_col or i_beg > im_ending_col:
                    continue
                # implement cosm_tail
                if i_beg+cosm_filter+cosm_tail+1 > mask.shape[2]:
                    ex_l = np.append(ex_l,
                            i_beg+cosm_filter+cosm_tail+1-mask.shape[2])
                    cutoffs = np.append(cutoffs, i+1)
                streak_end = int(min(i_beg+cosm_filter+cosm_tail+1,
                                mask.shape[2]))
                mask[j, i, i_beg:streak_end] = 1
                # implement cosm_box
                # can't have cosm_box appear in non-image pixels
                st_row = max(i-cosm_box, im_starting_row)
                end_row = min(i+cosm_box+1, im_ending_row+1)
                st_col = max(i_beg-cosm_box, im_starting_col)
                end_col = min(i_beg+cosm_box+1, im_ending_col+1)
                mask[j, st_row:end_row, st_col:end_col] = 1
                pass

        if mode == 'full' and len(ex_l) > 0:
            mask_rav = mask[j].ravel()
            for k in range(len(ex_l)):
                row = cutoffs[k]
                rav_ind = int(row * mask.shape[2] - 1)
                mask_rav[rav_ind:rav_ind + int(ex_l[k])] = 1


    return mask

def find_plateaus(streak_row, fwc, sat_thresh, plat_thresh, cosm_filter):
    """Find the beginning index of each cosmic plateau in a row.

    Args:
        streak_row (array_like, float):
            Row with possible cosmics.
        fwc (float):
            Full well capacity of detector *in DNs*.  Note that this may require a
            conversion as FWCs are usually specified in electrons, but the image
            is in DNs at this point.
        sat_thresh (float):
            Multiplication factor for fwc that determines saturated cosmic pixels.
        plat_thresh (float):
            Multiplication factor for fwc that determines edges of cosmic plateau.
        cosm_filter (float):
            Minimum length in pixels of cosmic plateaus to be identified.

    Returns:
        array_like, int:
            Index of plateau beginnings, or None if there is no plateau.
    """
    # Lowpass filter row to differentiate plateaus from standalone pixels
    # The way median_filter works, it will find cosmics that are cosm_filter-1
    # wide. Add 1 to cosm_filter to correct for this
    filtered = median_filter(streak_row, cosm_filter+1, mode='nearest')
    saturated = (filtered >= sat_thresh*fwc).nonzero()[0]

    if len(saturated) > 0:
        i_begs = np.array([])
        for i in range(len(saturated)):
            i_beg = saturated[i]
            while i_beg > 0 and streak_row[i_beg] >= plat_thresh*fwc:
                i_beg -= 1
            # unless saturated at col 0, shifts forward 1 to plateau start
            if streak_row[i_beg] < plat_thresh*fwc:
                i_beg += 1
            i_begs = np.append(i_begs, i_beg)

        return np.unique(i_begs).astype(int)
    else:
        return None
    
def calc_sat_fwc(emgain_arr,fwcpp_arr,fwcem_arr,sat_thresh):
	"""Calculates the lowest full well capacity saturation threshold for each frame.

	Args:
    	emgain_arr (np.array): 1D array of the EM gain value for each frame.
        fwcpp_arr (np.array): 1D array of the full-well capacity in the image
            frame (before em gain readout) value for each frame.
        fwcem_arr (np.array): 1D array of the full-well capacity in the EM gain
            register for each frame.
        sat_thresh (float): Multiplier for the full-well capacity to determine
            what qualifies as saturation. A reasonable value is 0.99

    Returns:
        np.array: lowest full well capacity saturation threshold for frames
    """
	possible_sat_fwcs_arr = np.append((emgain_arr * fwcpp_arr)[:,np.newaxis], fwcem_arr[:,np.newaxis],axis=1)
	sat_fwcs = sat_thresh * np.min(possible_sat_fwcs_arr,axis=1)

	return sat_fwcs

def nan_flags(dataset,threshold=1):
    """Replaces each DQ-flagged pixel (>= the given threshold) in the dataset with np.nan.

    Args:
        dataset (corgidrp.data.Dataset): input dataset.
        threshold (int, optional): DQ threshold to replace with nans. Defaults to 1.

    Returns:
        corgidrp.data.Dataset: dataset with flagged pixels replaced.
    """

    dataset_out = dataset.copy()

    # mask bad pixels
    bad = np.where(dataset_out.all_dq >= threshold)
    dataset_out.all_data[bad] = np.nan

    new_error = np.zeros_like(dataset_out.all_data)
    new_error[bad] = np.nan
    dataset_out.add_error_term(new_error, 'DQ flagged')
    
    return dataset_out

def flag_nans(dataset,flag_val=1):
    """Assigns a DQ flag to each nan pixel in the dataset.

    Args:
        dataset (corgidrp.data.Dataset): input dataset.
        flag_val (int, optional): DQ value to assign. Defaults to 1.

    Returns:
        corgidrp.data.Dataset: dataset with nan values flagged.
    """

    dataset_out = dataset.copy()

    # mask bad pixels
    bad = np.isnan(dataset_out.all_data)
    dataset_out.all_dq[bad] = flag_val

    return dataset_out
